import dataclasses
import warnings
from functools import partial
from typing import List, Optional, Tuple, NamedTuple, Any

import jax
import jax.numpy as jnp
import numpy as np
from jax import core
from jax._src.mesh import Mesh
from jax._src.partition_spec import PartitionSpec
from jax.experimental.shard_map import shard_map
from jaxlib import xla_client

from jaxns.framework.bases import BaseAbstractModel
from jaxns.internals.cumulative_ops import cumulative_op_static
from jaxns.internals.log_semiring import LogSpace, normalise_log_space
from jaxns.internals.maps import create_mesh, tree_device_put, replace_index
from jaxns.internals.mixed_precision import mp_policy
from jaxns.internals.random import sample_uniformly_masked, resample_indicies
from jaxns.internals.shrinkage_statistics import EvidenceUpdateVariables, _update_evidence_calc_op, \
    compute_evidence_stats
from jaxns.internals.stats import linear_to_log_stats, effective_sample_size_kish
from jaxns.internals.tree_structure import SampleTreeGraph, count_crossed_edges
from jaxns.internals.types import PRNGKey, IntArray, BoolArray
from jaxns.nested_samplers.abc import AbstractNestedSampler
from jaxns.nested_samplers.common.initialisation import create_init_state, create_init_termination_register
from jaxns.nested_samplers.common.termination import determine_termination
from jaxns.nested_samplers.common.types import TerminationCondition, NestedSamplerState, TerminationRegister, \
    SampleCollection, LivePointCollection, NestedSamplerResults
from jaxns.samplers.abc import AbstractSampler
from jaxns.samplers.abc import EphemeralState
from jaxns.samplers.uniform_samplers import UniformSampler

__all__ = [
    'ShardedStaticNestedSampler'
]


def _add_samples_to_state(sample_collection: LivePointCollection,
                          state: NestedSamplerState,
                          is_phantom: bool) -> NestedSamplerState:
    """
    Adds samples to state.

    Args:
        sample_collection: batched [N] samples
        state: state
        is_phantom: whether samples are phantom

    Returns:
        updated state
    """
    replace_idx = state.next_sample_idx
    num_samples = np.shape(sample_collection.log_L)[0]
    sample_collection = SampleCollection(
        sender_node_idx=replace_index(
            state.sample_collection.sender_node_idx,
            sample_collection.sender_node_idx,
            replace_idx
        ),
        log_L=replace_index(state.sample_collection.log_L, sample_collection.log_L, replace_idx),
        U_samples=replace_index(state.sample_collection.U_samples, sample_collection.U_sample, replace_idx),
        num_likelihood_evaluations=replace_index(
            state.sample_collection.num_likelihood_evaluations,
            sample_collection.num_likelihood_evaluations,
            replace_idx
        ),
        phantom=replace_index(
            state.sample_collection.phantom,
            jnp.full((num_samples,), is_phantom, jnp.bool_),
            replace_idx
        )
    )
    num_added = jnp.asarray(num_samples, mp_policy.index_dtype)
    next_sample_idx = state.next_sample_idx + num_added
    # Wrap at the number of samples (this is a trick for global optimisation that doesn't care about the entire progress)
    next_sample_idx = next_sample_idx % jnp.asarray(np.shape(state.sample_collection.log_L)[0], mp_policy.index_dtype)
    state = NestedSamplerState(
        key=state.key,
        next_sample_idx=next_sample_idx,
        sample_collection=sample_collection,
        num_samples=state.num_samples + num_added
    )
    return state


def get_samples(key, mesh: Mesh, sampler: AbstractSampler, sampler_state: Any, log_L_contour, num_samples: int):
    """
    Get samples from the sampler.

    Args:
        key: the PRNG key
        mesh: the mesh to use
        sampler: the sampler to use
        sampler_state: the sampler state
        log_L_contour: the log likelihood contour
        num_samples: the number of samples to get

    Returns:
        samples, phantom_samples
    """

    @partial(
        shard_map,
        mesh=mesh,
        in_specs=(PartitionSpec('shard', ), PartitionSpec(), PartitionSpec()),
        out_specs=(PartitionSpec('shard', ), PartitionSpec('shard', )),
        check_rep=False
    )
    def get_samples(sample_keys, log_L_contour, sampler_state: Any):
        def body(carry, sample_key):
            sample, phantom_samples = sampler.get_sample(
                key=sample_key,
                log_L_constraint=log_L_contour,
                sampler_state=sampler_state
            )
            return carry, (sample, phantom_samples)

        _, (sample, phantom_samples) = jax.lax.scan(body, None, sample_keys)
        # phantom samples are [k, ...] and samples are [...] so
        phantom_samples = jax.tree.map(
            lambda x: jax.lax.reshape(x, (int(np.prod(np.shape(x)[:2])),) + np.shape(x)[2:]),
            phantom_samples
        )
        return sample, phantom_samples

    sharded_sample_keys = tree_device_put(jax.random.split(key, num_samples), mesh, ('shard',))
    return get_samples(sharded_sample_keys, log_L_contour, sampler_state)


def dynamic_new_live_point_collection(
        key,
        mesh: Mesh,
        state: NestedSamplerState,
        sender_sample_idx: IntArray,
        sampler: AbstractSampler,
        sampler_state: Any,
        num_live_points: int
):
    """
    Get live points from the sampler for a dynamic slice.

    Args:
        mesh: the mesh to use
        state: the state
        sampler: the sampler
        sampler_state: the sampler state
        num_live_points: the number of live points to get

    Returns:
        live_point_collection, state
    """
    log_L_contour = state.sample_collection.log_L[sender_sample_idx]

    new_samples, phantom_samples = get_samples(
        key=key,
        mesh=mesh,
        sampler=sampler,
        sampler_state=sampler_state,
        log_L_contour=log_L_contour,
        num_samples=num_live_points
    )
    # Node is sample_idx - 1
    sender_node_idx = sender_sample_idx - jnp.asarray(1, mp_policy.index_dtype)
    live_point_collection = LivePointCollection(
        sender_node_idx=jnp.full((num_live_points,), sender_node_idx, mp_policy.index_dtype),
        U_sample=new_samples.U_sample,
        log_L=new_samples.log_L,
        log_L_constraint=new_samples.log_L_constraint,
        num_likelihood_evaluations=new_samples.num_likelihood_evaluations
    )

    sort_indices = jnp.argsort(live_point_collection.log_L)
    live_point_collection = jax.tree.map(lambda x: x[sort_indices], live_point_collection)

    state = add_phantom_samples_to_state(phantom_samples, sender_node_idx, state)
    return live_point_collection, state


def add_phantom_samples_to_state(phantom_samples, sender_node_idx, state):
    """
    Add phantom samples to the state

    Args:
        phantom_samples: the phantom samples
        sender_node_idx: the sender node index
        state: the state

    Returns:
        live_point_collection, state
    """
    # Add phantom samples (this is an option of user, controlled by `k`)
    num_phantom = np.shape(phantom_samples.log_L)[0]
    phantom_collection = LivePointCollection(
        sender_node_idx=jnp.full((num_phantom,), sender_node_idx, mp_policy.index_dtype),
        U_sample=phantom_samples.U_sample,
        log_L=phantom_samples.log_L,
        log_L_constraint=phantom_samples.log_L_constraint,
        num_likelihood_evaluations=phantom_samples.num_likelihood_evaluations
    )
    state = _add_samples_to_state(
        sample_collection=phantom_collection,
        state=state,
        is_phantom=True
    )
    return state


def _collect_shell(
        mesh: Mesh,
        live_point_collection: LivePointCollection,
        state: NestedSamplerState,
        termination_register: TerminationRegister,
        sampler: AbstractSampler,
        sampler_state: Any,
        shell_size: int
) -> Tuple[LivePointCollection, NestedSamplerState, TerminationRegister]:
    """
    Run nested sampling until `num_samples` samples are collected.

    Args:
        mesh: the device mesh to use
        live_point_collection: the live point collection
        state: the state of the nested sampler at the start
        termination_register: the termination register at the start
        sampler: sampler to use
        sampler_state: the sampler state to use for
        shell_size: the size of the shell to collect

    Returns:
        live_point_collection: the live point collection
        state: the state of the nested sampler at the end
        termination_register: the termination register at the end
    """

    # Find and discard shell
    front_size = np.shape(live_point_collection.log_L)[0]
    # always leave live points sorted so that we don't need to do it here.
    discarded_sample_collection: LivePointCollection = jax.tree.map(lambda x: x[:shell_size], live_point_collection)
    state = _add_samples_to_state(
        sample_collection=discarded_sample_collection,
        state=state,
        is_phantom=False
    )

    # Replace the discarded samples
    key, sample_key = jax.random.split(state.key, 2)
    state = state._replace(key=key)
    supremum_index = shell_size - 1  # Biggest of discarded
    log_L_contour = live_point_collection.log_L[supremum_index]
    new_samples, phantom_samples = get_samples(
        key=sample_key,
        mesh=mesh,
        sampler=sampler,
        sampler_state=sampler_state,
        log_L_contour=log_L_contour,
        num_samples=shell_size
    )
    # Sender is the maximum log_L sample from discarded, i.e. that last added discarded
    sender_node_idx = state.next_sample_idx - jnp.asarray(1, mp_policy.index_dtype)
    new_sample_collection = LivePointCollection(
        sender_node_idx=jnp.full((shell_size,), sender_node_idx, mp_policy.index_dtype),
        U_sample=new_samples.U_sample,
        log_L=new_samples.log_L,
        log_L_constraint=new_samples.log_L_constraint,
        num_likelihood_evaluations=new_samples.num_likelihood_evaluations
    )
    live_point_collection: LivePointCollection = jax.tree.map(
        lambda x, update: x.at[:shell_size, ...].set(update),
        live_point_collection,
        new_sample_collection
    )
    sort_indices = jnp.argsort(live_point_collection.log_L)
    live_point_collection = jax.tree.map(lambda x: x[sort_indices], live_point_collection)
    # TODO: compute insert index KS-statistic
    _, insert_indices = jax.lax.top_k(-sort_indices, k=shell_size)

    state = add_phantom_samples_to_state(phantom_samples, sender_node_idx, state)

    # Update termination register
    # Technically we must compute the num live points in case there is a plateau.
    #  but it gets done properly in to_results.
    evidence_calc, _ = cumulative_op_static(
        op=_update_evidence_calc_op,
        init=termination_register.evidence_calc,
        xs=EvidenceUpdateVariables(
            num_live_points=jnp.full((shell_size,), front_size, mp_policy.measure_dtype),
            log_L_next=discarded_sample_collection.log_L
        ),
    )
    evidence_calc_with_remaining, _ = cumulative_op_static(
        op=_update_evidence_calc_op,
        init=evidence_calc,
        xs=EvidenceUpdateVariables(
            num_live_points=jnp.arange(front_size, 0., -1., mp_policy.measure_dtype),
            log_L_next=live_point_collection.log_L
        ),
    )
    # Note we consider phantom samples requiring 0 num_likelihood_evaluations
    num_likelihood_evaluations = termination_register.num_likelihood_evaluations + jnp.sum(
        new_samples.num_likelihood_evaluations)
    # We determine efficiency
    efficiency = jnp.asarray(front_size / jnp.sum(live_point_collection.num_likelihood_evaluations),
                             mp_policy.measure_dtype)
    plateau = jnp.all(jnp.equal(live_point_collection.log_L, live_point_collection.log_L[0]))
    absolute_spread = jnp.abs(live_point_collection.log_L[-1] - live_point_collection.log_L[0])
    relative_spread = 2. * absolute_spread / jnp.abs(live_point_collection.log_L[0] + live_point_collection.log_L[-1])
    no_seed_points = live_point_collection.log_L[supremum_index] >= live_point_collection.log_L[-1]
    peak_log_XL = jnp.maximum(termination_register.peak_log_XL, evidence_calc.log_X_mean + evidence_calc.log_L)
    termination_register = TerminationRegister(
        num_samples_used=state.num_samples,
        evidence_calc=evidence_calc,
        evidence_calc_with_remaining=evidence_calc_with_remaining,
        num_likelihood_evaluations=num_likelihood_evaluations,
        log_L_contour=log_L_contour,
        efficiency=efficiency,
        plateau=plateau,
        no_seed_points=no_seed_points,
        relative_spread=relative_spread,
        absolute_spread=absolute_spread,
        peak_log_XL=peak_log_XL
    )
    return live_point_collection, state, termination_register


def _dynamic_posterior_refinement_iteration(
        mesh: Mesh,
        state: NestedSamplerState,
        sampler: AbstractSampler,
        termination_register: TerminationRegister,
        refine_threshold: float,
        num_live_points: int
) -> NestedSamplerState:
    """
    Perform a dynamic posterior refinement iteration.

    Args:
        mesh: the mesh to use
        state: the state of the nested sampler
        sampler: the sampler to use
        termination_register: the termination register
        refine_threshold: the threshold for refinement, will attach new points from somewhere within a masked region,
            where XL > max(XL) * refine_threshold
        num_live_points: the number of live points to get

    Returns:
        the updated state
    """
    key, sample_key, ephemeral_key, reseed_key, select_attach_key = jax.random.split(state.key, 5)
    state = state._replace(key=key)
    sample_collection = state.sample_collection

    # Get the part of prior region that contains most of the posterior mass.
    # This is where XL is largest.
    sample_tree = SampleTreeGraph(
        sender_node_idx=sample_collection.sender_node_idx,
        log_L=sample_collection.log_L
    )
    num_samples = jnp.minimum(
        state.num_samples,
        jnp.asarray(state.sample_collection.log_L.size, mp_policy.count_dtype)
    )
    live_point_counts = count_crossed_edges(sample_tree=sample_tree, num_samples=num_samples)
    num_live_points_per_sample = live_point_counts.num_live_points
    log_L = sample_tree.log_L[live_point_counts.samples_indices]
    final_evidence_stats, per_sample_evidence_stats = compute_evidence_stats(
        log_L=log_L,
        num_live_points=num_live_points_per_sample,
        num_samples=num_samples
    )
    log_XL = per_sample_evidence_stats.log_X_mean + per_sample_evidence_stats.log_L
    peak_log_XL = jnp.max(log_XL)

    # Randomly choose attach point
    # TODO: ensure attachment certain to be within sample set.
    select_mask = jnp.logical_and(
        log_XL > peak_log_XL + np.log(refine_threshold),
        jnp.logical_not(sample_collection.phantom)
    )
    log_weights = jnp.where(select_mask, 0., -jnp.inf)
    sample_root_idx = resample_indicies(select_attach_key, log_weights, 1)[0]

    # Create a fake live point collection to create sampler state. We don't care about the samples being i.i.d.
    log_L_constraint = sample_collection.log_L[sample_root_idx]
    select_mask = sample_collection.log_L > log_L_constraint
    ephemeral_sample_collection: SampleCollection = sample_uniformly_masked(
        key=reseed_key,
        v=sample_collection,
        select_mask=select_mask,
        num_samples=num_live_points
    )
    ephemeral_live_point_collection = LivePointCollection(
        sender_node_idx=ephemeral_sample_collection.sender_node_idx,
        U_sample=ephemeral_sample_collection.U_samples,
        log_L=ephemeral_sample_collection.log_L,
        log_L_constraint=sample_collection.log_L[ephemeral_sample_collection.sender_node_idx + 1],
        num_likelihood_evaluations=ephemeral_sample_collection.num_likelihood_evaluations
    )
    ephemeral_state = EphemeralState(
        key=ephemeral_key,
        live_points_collection=ephemeral_live_point_collection,
        termination_register=termination_register
    )
    sampler_state = sampler.pre_process(ephemeral_state)

    # Get a proper i.i.d. set of live points
    live_point_collection, state = dynamic_new_live_point_collection(
        key=sample_key,
        mesh=mesh,
        state=state,
        sender_sample_idx=sample_root_idx,
        sampler=sampler,
        sampler_state=sampler_state,
        num_live_points=num_live_points
    )

    # Add all live points to the state
    state = _add_samples_to_state(
        sample_collection=live_point_collection,
        state=state,
        is_phantom=False
    )
    return state


def _main_ns_thread(
        mesh: Mesh,
        live_point_collection: LivePointCollection,
        state: NestedSamplerState,
        termination_register: TerminationRegister,
        termination_cond: TerminationCondition,
        sampler: AbstractSampler,
        num_discards_per_iteration: int,
        shell_fraction: float,
        verbose: bool
) -> Tuple[LivePointCollection, NestedSamplerState, TerminationRegister, IntArray]:
    """
    Runs a single thread of static nested sampling until a stopping condition is reached. Discards 1/2 of the
    live points at once, replacing them from the supremum contour, creating a sample tree.

    Args:
        mesh: the device mesh to use
        state: the state of the nested sampler at the start
        termination_register: the termination register at the start
        termination_cond: the termination condition
        sampler: the sampler to use
        num_discards_per_iteration: number of discard shells per iteration, between processing sampler state.
        verbose: whether to log debug messages.

    Returns:
        live_point_collection: the final set of live points
        state: the final state
        termination_register: the termination register
        termination_condition: the reason for termination
    """
    if num_discards_per_iteration <= 0:
        raise ValueError("num_discards_per_iteration must be > 0 got {num_discards_per_iteration}.")

    # Update the termination condition to stop before going over the maximum number of samples.
    shell_size = int(np.shape(live_point_collection.log_L)[0] * shell_fraction)
    space_needed_per_iteration = num_discards_per_iteration * shell_size * (1 + sampler.num_phantom())

    if termination_cond.max_samples is not None:
        termination_cond = termination_cond._replace(
            max_samples=jnp.minimum(
                termination_cond.max_samples,
                np.shape(state.sample_collection.log_L)[0] - space_needed_per_iteration
            )
        )
    # Catch case of no seed points left
    no_seed_points = live_point_collection.log_L[shell_size - 1] >= live_point_collection.log_L[-1]
    termination_register = termination_register._replace(no_seed_points=no_seed_points)

    class CarryType(NamedTuple):
        live_point_collection: LivePointCollection
        state: NestedSamplerState
        termination_register: TerminationRegister

    def cond(carry: CarryType) -> BoolArray:
        done, termination_reason = determine_termination(
            term_cond=termination_cond,
            termination_register=carry.termination_register
        )
        return jnp.bitwise_not(done)

    def body(carry: CarryType) -> CarryType:
        # Discard half the live points and replace them with new samples
        live_point_collection, state, termination_register = carry

        key, ephemeral_key = jax.random.split(state.key, 2)
        state = state._replace(key=key)
        ephemeral_state = EphemeralState(
            key=ephemeral_key,
            live_points_collection=live_point_collection,
            termination_register=termination_register
        )
        sampler_state = sampler.pre_process(ephemeral_state)

        for _ in range(num_discards_per_iteration):
            live_point_collection, state, termination_register = _collect_shell(
                mesh=mesh,
                live_point_collection=live_point_collection,
                state=state,
                sampler=sampler,
                termination_register=termination_register,
                sampler_state=sampler_state,
                shell_size=shell_size
            )
            key, ephemeral_key = jax.random.split(state.key, 2)
            state = state._replace(key=key)
            ephemeral_state = EphemeralState(
                key=ephemeral_key,
                live_points_collection=live_point_collection,
                termination_register=termination_register
            )
            sampler_state = sampler.post_process(ephemeral_state=ephemeral_state, sampler_state=sampler_state)

        if verbose:
            log_Z_mean, log_Z_var = linear_to_log_stats(
                log_f_mean=termination_register.evidence_calc_with_remaining.log_Z_mean,
                log_f2_mean=termination_register.evidence_calc_with_remaining.log_Z2_mean)
            log_Z_uncert = jnp.sqrt(log_Z_var)
            log_Z_mean0, log_Z_var0 = linear_to_log_stats(
                log_f_mean=termination_register.evidence_calc.log_Z_mean,
                log_f2_mean=termination_register.evidence_calc.log_Z2_mean)
            log_Z_remaining = log_Z_mean - log_Z_mean0
            log_Z_remaining_error = jnp.sqrt(log_Z_var + log_Z_var0)
            ess = effective_sample_size_kish(termination_register.evidence_calc_with_remaining.log_Z_mean,
                                             termination_register.evidence_calc_with_remaining.log_dZ2_mean)
            jax.debug.print(
                "-------\n"
                "Num samples: {num_samples}\n"
                "Num likelihood evals: {num_likelihood_evals}\n"
                "Efficiency: {efficiency}\n"
                "log(L) contour: {log_L_contour}\n"
                "log(Z) est.: {log_Z_mean} +- {log_Z_uncert}\n"
                "log(Z | remaining) est.: {log_Z_remaining} +- {log_Z_remaining_error}\n"
                "ESS: {ess}\n",
                num_samples=termination_register.num_samples_used,
                num_likelihood_evals=termination_register.num_likelihood_evaluations,
                efficiency=termination_register.efficiency,
                log_L_contour=termination_register.log_L_contour,
                log_Z_mean=log_Z_mean,
                log_Z_uncert=log_Z_uncert,
                log_Z_remaining=log_Z_remaining,
                log_Z_remaining_error=log_Z_remaining_error,
                ess=ess
            )

        return CarryType(
            state=state,
            termination_register=termination_register,
            live_point_collection=live_point_collection
        )

    carry = CarryType(
        state=state,
        termination_register=termination_register,
        live_point_collection=live_point_collection
    )

    carry = jax.lax.while_loop(
        cond_fun=cond,
        body_fun=body,
        init_val=carry
    )

    _, termination_reason = determine_termination(
        term_cond=termination_cond,
        termination_register=carry.termination_register
    )

    return carry.live_point_collection, carry.state, carry.termination_register, termination_reason


def round_up_num_live_points(init_num_live_points, shell_frac, num_devices):
    num_live_points = int(init_num_live_points)
    while True:
        shell_size = int(num_live_points * shell_frac)
        if shell_size % num_devices == 0:
            break
        num_live_points += 1
    return num_live_points


def round_up_max_samples(init_max_samples, num_discard, num_phantom_points):
    max_samples = int(init_max_samples)
    block_size = num_discard * (1 + num_phantom_points)
    while True:
        if max_samples % block_size == 0:
            break
        max_samples += 1
    return max_samples


@dataclasses.dataclass(eq=False)
class ShardedStaticNestedSampler(AbstractNestedSampler):
    """
    A static nested sampler that uses a fixed number of live points. This uses a uniform sampler to generate the
    initial set of samples down to an efficiency threshold, then uses a provided sampler to generate the rest of the
    samples until the termination condition is met.

    Args:
        init_efficiency_threshold: the efficiency threshold to use for the initial uniform sampling. If 0 then
            turns it off.
        sampler: the sampler to use after the initial uniform sampling.
        num_live_points: the number of live points to use.
        model: the model to use.
        max_samples: the maximum number of samples to take.
        devices: the devices to use, default is 1.
        verbose: whether to log as we go.
    """
    model: BaseAbstractModel
    max_samples: int
    init_efficiency_threshold: float
    sampler: AbstractSampler
    num_live_points: int
    shell_fraction: Optional[float] = None
    num_dynamic_refinement_iterations: int = 0
    refine_threshold: float = 0.01
    devices: Optional[List[xla_client.Device]] = None
    verbose: bool = False

    def __post_init__(self):
        if self.shell_fraction is None:
            self.shell_fraction = 0.5
        self.shell_fraction = max(self.shell_fraction, 1. / self.num_live_points)
        if (self.shell_fraction <= 0.) or (self.shell_fraction > 1.):
            raise ValueError(
                f"Expected 0 < shell_fraction <= 1, got {self.shell_fraction}. Best to keep it around 0.5.")
        if self.devices is None:
            self.devices = jax.devices()
        if len(self.devices) > 1:
            print(f"Running over {len(self.devices)} devices.")
        # Make sure num_live_points // 2 is a multiple of the number of devices
        self.num_live_points = round_up_num_live_points(
            init_num_live_points=self.num_live_points,
            shell_frac=self.shell_fraction,
            num_devices=len(self.devices)
        )
        self.max_samples = round_up_max_samples(
            init_max_samples=self.max_samples,
            # TODO: if we do more than 1 discard per iteration need to update here too.
            num_discard=int(self.shell_fraction * self.num_live_points),
            num_phantom_points=self.sampler.num_phantom()
        )

        if self.num_dynamic_refinement_iterations > 0:
            warnings.warn("Dynamic refinement is experimental and may not work as expected.")

    def _to_results(self, termination_reason: IntArray, state: NestedSamplerState,
                    trim: bool) -> NestedSamplerResults:

        num_samples = jnp.minimum(
            state.num_samples,
            jnp.asarray(state.sample_collection.log_L.size, mp_policy.count_dtype)
        )

        sample_collection = state.sample_collection

        if trim:
            trim_size = jnp.minimum(
                state.num_samples,
                jnp.asarray(state.sample_collection.log_L.size, mp_policy.count_dtype)
            )

            if isinstance(num_samples, core.Tracer):
                raise RuntimeError("Tracer detected, but expected imperative context.")
            sample_collection = jax.tree.map(lambda x: x[:trim_size], sample_collection)

            sample_tree = SampleTreeGraph(
                sender_node_idx=sample_collection.sender_node_idx,
                log_L=sample_collection.log_L
            )

            live_point_counts = count_crossed_edges(sample_tree=sample_tree)
            num_live_points = live_point_counts.num_live_points
            log_L = sample_tree.log_L[live_point_counts.samples_indices]
            U_samples = sample_collection.U_samples[live_point_counts.samples_indices]
            num_likelihood_evaluations = sample_collection.num_likelihood_evaluations[
                live_point_counts.samples_indices]
            final_evidence_stats, per_sample_evidence_stats = compute_evidence_stats(
                log_L=log_L,
                num_live_points=num_live_points
            )
        else:
            sample_tree = SampleTreeGraph(
                sender_node_idx=sample_collection.sender_node_idx,
                log_L=sample_collection.log_L
            )

            live_point_counts = count_crossed_edges(sample_tree=sample_tree, num_samples=num_samples)
            num_live_points = live_point_counts.num_live_points
            log_L = sample_tree.log_L[live_point_counts.samples_indices]
            U_samples = sample_collection.U_samples[live_point_counts.samples_indices]
            num_likelihood_evaluations = sample_collection.num_likelihood_evaluations[
                live_point_counts.samples_indices]

            final_evidence_stats, per_sample_evidence_stats = compute_evidence_stats(
                log_L=log_L,
                num_live_points=num_live_points,
                num_samples=num_samples
            )

        log_Z_mean, log_Z_var = linear_to_log_stats(
            log_f_mean=final_evidence_stats.log_Z_mean,
            log_f2_mean=final_evidence_stats.log_Z2_mean
        )
        log_Z_uncert = jnp.sqrt(log_Z_var)

        # Correction by sqrt(k+1)
        total_phantom_samples = jnp.sum(mp_policy.cast_to_count(sample_collection.phantom, quiet=True))
        phantom_fraction = total_phantom_samples / num_samples  # k / (k+1)
        k = phantom_fraction / (1. - phantom_fraction)
        log_Z_uncert = log_Z_uncert * jnp.sqrt(1. + k)

        # Kish's ESS = [sum dZ]^2 / [sum dZ^2]
        ESS = effective_sample_size_kish(final_evidence_stats.log_Z_mean, final_evidence_stats.log_dZ2_mean)
        ESS = ESS / (1. + k)

        samples = jax.vmap(self.model.transform)(U_samples)
        parametrised_samples = jax.vmap(self.model.transform_parametrised)(U_samples)

        log_L_samples = log_L
        dp_mean = LogSpace(per_sample_evidence_stats.log_dZ_mean)
        dp_mean = normalise_log_space(dp_mean)
        H_mean_instable = -(
                (
                        dp_mean * LogSpace.from_signed_value(
                    jnp.where(jnp.isneginf(dp_mean.log_abs_val), 0., log_L_samples)
                )
                ).sum().value - log_Z_mean
        )
        # H \approx E[-log(compression)] = E[-log(X)] (More stable than E[log(L) - log(Z)] but biased)
        H_mean_stable = -((dp_mean * LogSpace(jnp.log(-per_sample_evidence_stats.log_X_mean))).sum().value)
        H_mean = jnp.where(jnp.isfinite(H_mean_instable), H_mean_instable, H_mean_stable)
        X_mean = LogSpace(per_sample_evidence_stats.log_X_mean)
        num_likelihood_evaluations_per_sample = num_likelihood_evaluations
        total_num_likelihood_evaluations = jnp.sum(num_likelihood_evaluations_per_sample)
        num_live_points_per_sample = num_live_points
        efficiency = LogSpace(jnp.log(num_samples) - jnp.log(total_num_likelihood_evaluations))

        log_posterior_density = log_L + jax.vmap(self.model.log_prob_prior)(
            U_samples)

        return NestedSamplerResults(
            log_Z_mean=log_Z_mean,  # estimate of log(E[Z])
            log_Z_uncert=log_Z_uncert,  # estimate of log(StdDev[Z])
            ESS=ESS,  # estimate of Kish's effective sample size
            H_mean=H_mean,  # estimate of E[int log(L) L dp/Z]
            total_num_samples=num_samples,  # int, the total number of samples collected.
            total_phantom_samples=total_phantom_samples,  # int, the total number of phantom samples collected.
            log_L_samples=log_L_samples,  # log(L) of each sample
            log_dp_mean=dp_mean.log_abs_val,
            log_posterior_density=log_posterior_density,
            # log(E[dZ]) of each sample, where dZ is how much it contributes to the total evidence.
            # log(StdDev[dZ]) of each sample, where dZ is how much it contributes to the total evidence.
            log_X_mean=X_mean.log_abs_val,  # log(E[U]) of each sample
            num_likelihood_evaluations_per_sample=num_likelihood_evaluations_per_sample,
            # how many likelihood evaluations were made per sample.
            num_live_points_per_sample=num_live_points_per_sample,
            # how many live points were taken for the samples.
            total_num_likelihood_evaluations=total_num_likelihood_evaluations,
            # how many likelihood evaluations were made in total,
            # sum of num_likelihood_evaluations_per_sample.
            log_efficiency=efficiency.log_abs_val,
            # total_num_samples / total_num_likelihood_evaluations
            termination_reason=termination_reason,  # termination condition as bit mask
            samples=samples,
            parametrised_samples=parametrised_samples,
            U_samples=U_samples
        )

    def _run(self, key: PRNGKey, term_cond: TerminationCondition) -> Tuple[
        IntArray, TerminationRegister, NestedSamplerState]:
        # Create sampler threads.
        mesh = create_mesh((len(self.devices),), ('shard',), devices=self.devices)

        if self.verbose:
            jax.debug.print(f"Creating initial state with {self.num_live_points} live points.")

        live_point_collection, state = create_init_state(
            key=key,
            num_live_points=self.num_live_points,
            max_samples=self.max_samples,
            model=self.model,
            mesh=mesh
        )

        termination_register = create_init_termination_register()

        if self.init_efficiency_threshold > 0.:
            if self.verbose:
                jax.debug.print(
                    f"Running uniform sampling down to efficiency threshold of {self.init_efficiency_threshold}."
                )
            # Uniform sampling down to a given mean efficiency
            uniform_sampler = UniformSampler(model=self.model)
            uniform_term_cond = TerminationCondition(
                efficiency_threshold=jnp.asarray(self.init_efficiency_threshold, mp_policy.measure_dtype),
                dlogZ=jnp.asarray(0., mp_policy.measure_dtype),
                max_samples=jnp.asarray(self.max_samples, mp_policy.count_dtype)
            )
            live_point_collection, state, termination_register, termination_reason = _main_ns_thread(
                mesh=mesh,
                live_point_collection=live_point_collection,
                state=state,
                termination_register=termination_register,
                termination_cond=uniform_term_cond,
                sampler=uniform_sampler,
                num_discards_per_iteration=1,
                shell_fraction=self.shell_fraction,
                verbose=self.verbose
            )
        if self.verbose:
            jax.debug.print("Running until termination condition: {term_cond}",
                            term_cond=term_cond)

        # Continue sampling with provided sampler until user-defined termination condition is met.
        live_point_collection, state, termination_register, termination_reason = _main_ns_thread(
            mesh=mesh,
            live_point_collection=live_point_collection,
            state=state,
            termination_register=termination_register,
            termination_cond=term_cond,
            sampler=self.sampler,
            num_discards_per_iteration=1,
            shell_fraction=self.shell_fraction,
            verbose=self.verbose
        )

        # Consumer live_point_collection
        state = _add_samples_to_state(
            sample_collection=live_point_collection,
            state=state,
            is_phantom=False
        )

        def body(i, state: NestedSamplerState) -> NestedSamplerState:
            state = _dynamic_posterior_refinement_iteration(
                mesh=mesh, state=state, sampler=self.sampler, termination_register=termination_register,
                refine_threshold=self.refine_threshold,
                num_live_points=self.num_live_points
            )
            return state

        if self.num_dynamic_refinement_iterations > 0:
            state = jax.lax.fori_loop(0, self.num_dynamic_refinement_iterations, body, state)

        return termination_reason, termination_register, state



from jaxns.samplers.multi_ellipsoidal_samplers import *
from jaxns.samplers.multi_slice_sampler import *
from jaxns.samplers.uni_slice_sampler import *
from jaxns.samplers.uniform_samplers import *
from jaxns.nested_samplers import *



import dataclasses
import warnings
from typing import TypeVar

import jax
from jax import numpy as jnp

from jaxns.internals.logging import get_grandparent_info

# We set this here also, since mp_policy use will mandate this being invoked early enough to matter.
jax.config.update("jax_threefry_partitionable", True)

if not jax.config.read('jax_enable_x64'):
    warnings.warn("JAX x64 is not enabled. Setting it now. Check for errors.")
    jax.config.update('jax_enable_x64', True)

# Create a float scalar to lock in dtype choices.
if jnp.array(1., jnp.float64).dtype != jnp.float64:
    raise RuntimeError("Failed to set float64 as default dtype.")

T = TypeVar("T")


def _cast_floating_to(tree: T, dtype: jnp.dtype, quiet: bool) -> T:
    def conditional_cast(x):
        if isinstance(x, float):
            return jnp.asarray(x, dtype=dtype)
        try:
            if not quiet and not jnp.issubdtype(x.dtype, jnp.floating):
                warnings.warn(f"Expected float type, got {x.dtype}, {get_grandparent_info()}.")
            return x.astype(dtype)
        except AttributeError:
            warnings.warn(f"Failed to cast {x} to {dtype}.")
            return x

    return jax.tree.map(conditional_cast, tree)


def _cast_complex_to(tree: T, dtype: jnp.dtype, quiet: bool) -> T:
    def conditional_cast(x):
        if isinstance(x, complex):
            return jnp.asarray(x, dtype=dtype)
        try:
            if not quiet and not jnp.issubdtype(x.dtype, jnp.complexfloating):
                warnings.warn(f"Expected complex type, got {x.dtype}, {get_grandparent_info()}.")
            return x.astype(dtype)
        except AttributeError:
            warnings.warn(f"Failed to cast {x} to {dtype}.")
            return x

    return jax.tree.map(conditional_cast, tree)


def _cast_integer_to(tree: T, dtype: jnp.dtype, quiet: bool) -> T:
    def conditional_cast(x):
        if isinstance(x, int):
            return jnp.asarray(x, dtype=dtype)
        try:
            if not quiet and not jnp.issubdtype(x.dtype, jnp.integer):
                warnings.warn(f"Expected integer type, got {x.dtype}, {get_grandparent_info()}.")
            return x.astype(dtype)
        except AttributeError:
            warnings.warn(f"Failed to cast {x} to {dtype}.")
            return x

    return jax.tree.map(conditional_cast, tree)


def _cast_bool_to(tree: T, dtype: jnp.dtype, quiet: bool) -> T:
    def conditional_cast(x):
        if isinstance(x, bool):
            return jnp.asarray(x, dtype=dtype)
        try:
            if not quiet and not jnp.issubdtype(x.dtype, jnp.bool_):
                warnings.warn(f"Expected bool type, got {x.dtype}, {get_grandparent_info()}.")
            return x.astype(dtype)
        except AttributeError:
            warnings.warn(f"Failed to cast {x} to {dtype}.")
            return x

    return jax.tree.map(conditional_cast, tree)


X = TypeVar("X")


@dataclasses.dataclass(frozen=True)
class Policy:
    """Encapsulates casting for inputs, outputs and parameters."""
    measure_dtype: jnp.dtype = jnp.float64
    index_dtype: jnp.dtype = jnp.int64
    count_dtype: jnp.dtype = jnp.int64

    def cast_to_index(self, x: X, quiet: bool = False) -> X:
        """Converts index values to the index dtype."""
        return _cast_integer_to(x, self.index_dtype, quiet=quiet)

    def cast_to_measure(self, x: X, quiet: bool = False) -> X:
        """Converts measure values to the measure dtype."""
        return _cast_floating_to(x, self.measure_dtype, quiet=quiet)

    def cast_to_count(self, x: X, quiet: bool = False) -> X:
        """Converts count values to the count dtype."""
        return _cast_integer_to(x, self.count_dtype, quiet=quiet)


mp_policy = Policy()

float_type = jnp.result_type(float)
int_type = jnp.result_type(int)
complex_type = jnp.result_type(complex)



import numpy as np
import pytest
from jax import numpy as jnp

from jaxns.experimental.solvers.gauss_newton_cg import newton_cg_solver


def test_newton_cg_solver():
    """A simple test for the newton_cg_solver function."""

    def obj_fn(x):
        return jnp.cos(jnp.sum(x ** 2))  # simple quadratic function

    x0 = jnp.array([1.0, 2.0, 3.0])  # initial guess
    x_final, diag = newton_cg_solver(obj_fn, x0, maxiter=10, verbose=True)

    print("Final solution:", x_final)
    print("Diagnostics:", diag)


@pytest.mark.parametrize("n", [1000])
def test_approx_cg_newton(n):
    def rosenbrock_nd(x):
        a = 1.
        b = 100.
        return jnp.sum((a - x[:-1]) ** 2 + b * (x[1:] - x[:-1] ** 2) ** 2)

    x0 = 10 * jnp.ones(n)

    solution, diagnostics = newton_cg_solver(rosenbrock_nd, x0, verbose=True)
    np.testing.assert_allclose(solution, jnp.ones(n), atol=1e-4)
    import pylab as plt
    plt.plot(np.log(diagnostics.g_norm), label="error")
    plt.plot(np.log(diagnostics.ddelta_x_norm), label="delta_norm")
    plt.plot(np.log(diagnostics.g_norm / diagnostics.ddelta_x_norm), label="error/delta_norm")
    plt.plot(np.log(diagnostics.damping), label="damping")
    plt.plot(np.log(diagnostics.mu), label="mu")
    plt.legend()
    plt.show()



import warnings
from typing import Callable

import jax
import numpy as np

from jaxns.internals.mixed_precision import mp_policy
from jaxns.internals.types import LikelihoodType

__all__ = [
    'jaxify_likelihood'
]


def jaxify_likelihood(log_likelihood: Callable[..., np.ndarray], vectorised: bool = False) -> LikelihoodType:
    """
    Wraps a non-JAX log likelihood function.

    Args:
        log_likelihood: a non-JAX log-likelihood function, which accepts a number of arguments and returns a scalar
            log-likelihood.
        vectorised: if True then the `log_likelihood` must handle batched inputs, i.e. each input will receive a common
            set of batched dimensions which the function must handle.

    Returns:
        A JAX-compatible log-likelihood function.
    """
    warnings.warn(
        "You're using a non-JAX log-likelihood function. This may be slower than a JAX log-likelihood function. "
        "Also, you are responsible for ensuring that the function is deterministic. "
        "Also, you cannot use learnable parameters in the likelihood call."
    )

    def _cond_cast(x):
        if isinstance(x, (jax.Array, np.ndarray)):
            return np.asarray(x)
        return x

    def _casted_log_likelihood(*args) -> np.ndarray:
        args = jax.tree.map(_cond_cast, args)  # Convert all arguments to numpy arrays, as they now pass jax.Array
        return mp_policy.cast_to_measure(log_likelihood(*args))

    def _log_likelihood(*args) -> jax.Array:
        # Define the expected shape & dtype of output.
        result_shape_dtype = jax.ShapeDtypeStruct(
            shape=(),
            dtype=mp_policy.measure_dtype
        )
        if vectorised:
            return jax.pure_callback(_casted_log_likelihood, result_shape_dtype, *args, vmap_method='broadcast_all')
        return jax.pure_callback(_casted_log_likelihood, result_shape_dtype, *args, vmap_method=None)

    return _log_likelihood



from typing import NamedTuple, Callable, TypeVar, Tuple

import jax
import jax.numpy as jnp
import numpy as np

from jaxns.experimental.solvers.ad_utils import tree_vdot_real_part, tree_scalar_mul, tree_add, tree_sub
from jaxns.internals.types import IntArray, FloatArray


def _identity(x):
    return x


class CGDiagnostics(NamedTuple):
    iterations: IntArray
    final_res_norm: FloatArray


DomainType = TypeVar('DomainType')


def cg_solve(A: Callable[[DomainType], DomainType], b: DomainType, x0: DomainType,
             M: Callable[[DomainType], DomainType] = _identity, maxiter: int | None = 100, tol: float = 1e-5,
             atol: float = 0.0) -> Tuple[DomainType, CGDiagnostics]:
    """
    Solve a linear system Ax = b using the conjugate gradient method.

    Args:
        A: a square PSD linear operator
        b: the right-hand side
        x0: an initial guess for the solution
        M: a preconditioner for A
        maxiter: the maximum number of iterations, if None then size of b
        tol: the relative tolerance for the residual norm
        atol: the absolute tolerance for the residual norm

    Returns:
        the solution x and diagnostics
    """

    if maxiter is None:
        maxiter = sum(jax.tree.leaves(jax.tree.map(np.size, b)))

    class CGState(NamedTuple):
        x: DomainType  # current solution estimate
        r: DomainType  # residual: r = b - A(x)
        z: DomainType  # preconditioned residual: z = M(r)
        p: DomainType  # search direction
        gamma: FloatArray  # inner product gamma = <r, z>
        iter: IntArray  # current iteration count
        atol2: FloatArray  # squared absolute tolerance for residual norm

    def _create_initial_state(x0: DomainType) -> CGState:
        b_norm_sq = tree_vdot_real_part(b, b)
        atol2 = jnp.maximum(tol ** 2 * b_norm_sq, atol ** 2)
        r = tree_sub(b, A(x0))
        z = M(r)
        p = z
        dtype = jnp.result_type(*jax.tree.leaves(p))
        gamma = tree_vdot_real_part(r, z).astype(dtype)
        return CGState(x=x0, r=r, z=z, p=p, gamma=gamma, iter=0, atol2=atol2)

    def cond_fn(state: CGState):
        res_norm_sq = state.gamma.real if M == _identity else tree_vdot_real_part(state.r, state.r)
        return (res_norm_sq > state.atol2) & (state.iter < maxiter)

    def body_fn(state: CGState):
        # Cache A(p)
        q = A(state.p)
        dtype = jnp.result_type(*jax.tree.leaves(state.p))
        alpha = state.gamma / tree_vdot_real_part(state.p, q).astype(dtype)
        x_new = tree_add(state.x, tree_scalar_mul(alpha, state.p))
        r_new = tree_sub(state.r, tree_scalar_mul(alpha, q))
        z_new = M(r_new)
        gamma_new = tree_vdot_real_part(r_new, z_new).astype(dtype)
        beta = gamma_new / state.gamma
        p_new = tree_add(z_new, tree_scalar_mul(beta, state.p))
        return CGState(x=x_new, r=r_new, z=z_new, p=p_new,
                       gamma=gamma_new, iter=state.iter + 1, atol2=state.atol2)

    init_state = _create_initial_state(x0=x0)
    final_state = jax.lax.while_loop(cond_fn, body_fn, init_state)
    final_res_sq = final_state.gamma.real if M == _identity else tree_vdot_real_part(final_state.r, final_state.r)

    diag = CGDiagnostics(iterations=final_state.iter,
                         final_res_norm=jnp.sqrt(final_res_sq))
    return final_state.x, diag



import dataclasses
from typing import Optional

from jaxlib import xla_client

from jaxns.experimental import SimpleGlobalOptimisation, GlobalOptimisationTerminationCondition, \
    GlobalOptimisationResults
from jaxns.experimental.global_optimisation import plot_progress, go_summary
from jaxns.framework.bases import BaseAbstractModel
from jaxns.internals.types import PRNGKey
from jaxns.samplers import UniDimSliceSampler

__all__ = [
    'GlobalOptimisation',
    'DefaultGlobalOptimisation'
]


@dataclasses.dataclass(eq=False)
class GlobalOptimisation:
    """
    A global optimiser using nested sampling as the core algorithm. Can easily globally optimise complex models, with
    curving degeneracies and multimodal structure. Highly parallelisable. Recommended to use gradient information by
    setting gradient_slice=True.

    Note, the log-likelihood over the model is maximised NOT the posterior. The prior acts as the search space prior,
    by constraining the search space and giving search preference to regions of high prior probability. Thus, the
    prior should encode your prior belief about where you think the global maximum is located.

    Args:
        model: a model to perform global optimisation on over the sample space.
        num_search_chains: number of search chains to use.
        s: number of slices to use per dimension.
        k: number of phantom samples to use.
        gradient_slice: if true use gradient information to improve. Default True.
        shell_frac: fraction of the shell to discard in parallel.
        devices: devices to use for parallel sharded computation. Default all available devices.
        verbose: whether to print verbose output. Default False.
    """
    model: BaseAbstractModel
    num_search_chains: Optional[int] = None
    s: Optional[int] = None
    k: Optional[int] = None
    gradient_slice: bool = True
    shell_frac: Optional[float] = None
    devices: Optional[xla_client.Device] = None
    verbose: bool = False

    def __post_init__(self):
        if self.num_search_chains is None:
            if self.gradient_slice:
                self.num_search_chains = self.model.U_ndims * 15
            else:
                self.num_search_chains = self.model.U_ndims * 100
        if self.s is None:
            if self.gradient_slice:
                self.s = 2
            else:
                self.s = 10
        if self.shell_frac is None:
            if self.gradient_slice:
                self.shell_frac = 0.5
            else:
                self.shell_frac = 0.5
        if self.k is None:
            if self.gradient_slice:
                self.k = self.model.U_ndims * self.s - 1
            else:
                self.k = self.model.U_ndims * self.s - 1

        sampler = UniDimSliceSampler(
            model=self.model,
            num_slices=self.model.U_ndims * int(self.s),
            num_phantom_save=int(self.k),
            midpoint_shrink=True,
            perfect=True,
            gradient_slice=self.gradient_slice
        )

        self._global_optimiser = SimpleGlobalOptimisation(
            sampler=sampler,
            num_search_chains=int(self.num_search_chains),
            shell_frac=float(self.shell_frac),
            model=self.model,
            devices=self.devices,
            verbose=self.verbose
        )

        self.summary = go_summary
        self.plot_progress = plot_progress

    def __call__(self, key: PRNGKey,
                 term_cond: Optional[GlobalOptimisationTerminationCondition] = None,
                 finetune: bool = False) -> GlobalOptimisationResults:
        """
        Runs the global optimisation.

        Args:
            key: PRNGKey
            term_cond: termination condition
            finetune: whether to use gradient-based fine-tune. Default False because not all models have gradients.

        Returns:
            results of the global optimisation
        """
        if term_cond is None:
            term_cond = GlobalOptimisationTerminationCondition(
                min_efficiency=3e-2
            )
        termination_reason, state = self._global_optimiser._run(key, term_cond)
        results = self._global_optimiser._to_results(termination_reason, state)
        if finetune:
            results = self._global_optimiser._gradient_descent(results=results)
        return results


DefaultGlobalOptimisation = GlobalOptimisation



from jax import numpy as jnp

from jaxns.experimental.solvers.cg import cg_solve


def test_cg():
    A = jnp.array([[1.0, 0.0], [0.0, 1.0]])
    b = jnp.array([1.0, 1.0])
    x0 = jnp.array([0.0, 0.0])

    def A_op(x):
        return jnp.dot(A, x)

    x, diag = cg_solve(A_op, b, x0)
    assert jnp.allclose(x, b)
    assert diag.iterations == 1



from typing import Optional
from uuid import uuid4

import jax.tree
import numpy as np
from jax import random, vmap, jit, numpy as jnp

import jaxns.framework.context as ctx
from jaxns.framework.bases import BaseAbstractModel, PriorModelType
from jaxns.framework.ops import transform, prepare_input, compute_log_prob_prior, compute_log_likelihood, parse_prior, \
    parse_joint, transform_parametrised
from jaxns.internals.logging import logger
from jaxns.internals.maps import pytree_unravel
from jaxns.internals.mixed_precision import mp_policy
from jaxns.internals.types import PRNGKey, FloatArray, LikelihoodType, UType, XType, LikelihoodInputType, \
    WType

__all__ = [
    'Model'
]

# TODO: Rewrite Model using context framework to avoid recompiling functions.

class Model(BaseAbstractModel):
    """
    Represents a Bayesian model in terms of a generative prior, and likelihood function.
    """

    def __init__(self, prior_model: PriorModelType, log_likelihood: LikelihoodType,
                 params: Optional[ctx.MutableParams] = None):
        super().__init__(prior_model=prior_model, log_likelihood=log_likelihood)
        if params is None:
            params = self.init_params(rng=random.PRNGKey(0))
        self._params = params
        self._num_params = int(sum(jax.tree.leaves(jax.tree.map(np.size, self._params))))
        # Parse the prior model to get place holders
        self.__U_placeholder, self.__X_placeholder, self.__W_placeholder = ctx.transform(
            lambda: parse_prior(prior_model=self.prior_model)
        ).apply(self._params, random.PRNGKey(0)).fn_val
        self._id = str(uuid4())  # Used for making sure it's hashable, so it can be used as a key in a dict.
        self.ravel_fn, self.unravel_fn = pytree_unravel(self.__W_placeholder)

    def __repr__(self):
        return f"Model(U_ndims={self.U_ndims}, num_params={self.num_params})"

    @property
    def num_params(self) -> int:
        if self._params is None:
            raise RuntimeError("Model has not been initialised")
        leafs = jax.tree.leaves(self._params)
        leaf_sizes = jax.tree.map(np.size, leafs)
        return sum(leaf_sizes)

    @property
    def params(self):
        if self._params is None:
            raise RuntimeError("Model has not been initialised")
        return self._params

    def set_params(self, params: ctx.MutableParams) -> 'Model':
        """
        Create a new parametrised model with the given parameters.

        Args:
            params: The parameters to use.

        Returns:
            A model with set parameters.
        """
        return Model(prior_model=self.prior_model, log_likelihood=self.log_likelihood, params=params)

    def __call__(self, params: ctx.MutableParams) -> 'Model':
        """
        Create a new parametrised model with the given parameters.

        **This is (and must be) a pure function.**

        Args:
            params: The parameters to use.

        Returns:
            A model with set parameters.
        """
        return self.set_params(params=params)

    def _U_placeholder(self) -> UType:
        return self.__U_placeholder

    def _X_placeholder(self) -> XType:
        return self.__X_placeholder

    def _W_placeholder(self) -> WType:
        return self.__W_placeholder

    def init_params(self, rng: PRNGKey) -> ctx.MutableParams:
        """
        Initialise the parameters of the model.

        Args:
            rng: PRNGkey to initialise the parameters.

        Returns:
            The initialised parameters.
        """

        def _log_prob_joint():
            U_placeholder, X_placeholder, W_placeholder, likelihood_input_placeholder, log_L_placeholder = parse_joint(
                prior_model=self.prior_model,
                log_likelihood=self.log_likelihood
            )
            log_prob_prior = compute_log_prob_prior(W=W_placeholder, prior_model=self.prior_model)
            log_prob_likelihood = compute_log_likelihood(W=W_placeholder, prior_model=self.prior_model,
                                                         log_likelihood=self.log_likelihood, allow_nan=True)
            return log_prob_prior + log_prob_likelihood

        params = ctx.transform(_log_prob_joint).init(rng).params
        return params

    def __hash__(self):
        return hash(self._id)

    def sample_U(self, key: PRNGKey) -> UType:
        """
        Sample from the prior model.

        Args:
            key: PRNGKey to use.

        Returns:
            The sampled U.
        """
        if self._params is None:
            raise RuntimeError("Model has not been initialised")

        def _sample_U():
            return random.uniform(key=ctx.next_rng_key(), shape=(self.U_ndims,), dtype=mp_policy.measure_dtype)

        return ctx.transform(_sample_U).apply(self._params, key).fn_val

    def sample_W(self, key: PRNGKey) -> WType:
        """
        Sample from the prior model.

        Args:
            key: PRNGKey to use.

        Returns:
            The sampled W.
        """
        if self._params is None:
            raise RuntimeError("Model has not been initialised")

        return self.unravel_fn(self.sample_U(key=key))

    def transform(self, U: UType) -> XType:
        def _transform():
            return transform(W=self.unravel_fn(U), prior_model=self.prior_model)

        return ctx.transform(_transform).apply(self._params, random.PRNGKey(0)).fn_val

    def transform_parametrised(self, U: UType) -> XType:
        def _transform():
            return transform_parametrised(W=self.unravel_fn(U), prior_model=self.prior_model)

        return ctx.transform(_transform).apply(self._params, random.PRNGKey(0)).fn_val

    def forward(self, U: UType, allow_nan: bool = False) -> FloatArray:
        if self._params is None:
            raise RuntimeError("Model has not been initialised")

        def _forward():
            return compute_log_likelihood(W=self.unravel_fn(U), prior_model=self.prior_model,
                                          log_likelihood=self.log_likelihood,
                                          allow_nan=allow_nan)

        return ctx.transform(_forward).apply(self._params, random.PRNGKey(0)).fn_val

    def log_prob_prior(self, U: UType) -> FloatArray:
        if self._params is None:
            raise RuntimeError("Model has not been initialised")

        def _log_prob_prior():
            return compute_log_prob_prior(W=self.unravel_fn(U), prior_model=self.prior_model)

        return ctx.transform(_log_prob_prior).apply(self._params, random.PRNGKey(0)).fn_val

    def prepare_input(self, U: UType) -> LikelihoodInputType:
        if self._params is None:
            raise RuntimeError("Model has not been initialised")

        def _prepare_input():
            return prepare_input(W=self.unravel_fn(U), prior_model=self.prior_model)

        return ctx.transform(_prepare_input).apply(self._params, random.PRNGKey(0)).fn_val

    def sanity_check(self, key: PRNGKey, S: int):
        U = jit(vmap(self.sample_U))(random.split(key, S))
        log_L = jit(vmap(lambda u: self.forward(u, allow_nan=True)))(U)
        logger.info("Sanity check...")
        for _U, _log_L in zip(U, log_L):
            if jnp.isnan(_log_L):
                logger.info(f"Found bad point:"
                            f"\n{_U} -> {self.transform(_U)}"
                            f"\n -> {self.transform_parametrised(_U)}")
        assert not any(np.isnan(log_L))
        logger.info("Sanity check passed")
        if 'parsed_prior' in self.__dict__:
            del self.__dict__['parsed_prior']



import jax
import numpy as np
from jax import numpy as jnp

from jaxns.experimental.solvers.ad_utils import grad_and_hvp, build_hvp


def test_grad_and_hvp():
    def f(params):
        W, b = params
        x = jnp.ones_like(b)
        return jnp.sum(jax.nn.sigmoid(W @ x + b))

    W = jnp.array([[1., 2.], [3., 4.]])
    b = jnp.array([1., 2.])

    params = (W, b)

    v = params

    grad, hvp = grad_and_hvp(f, params, v)

    print(grad)
    print(hvp)


def test_build_hvp():
    def f_crazy(x):
        return jnp.sum(jnp.cos(x) ** 2) + jnp.sum(jnp.sin(x) ** 2)

    x = jnp.ones((10,))
    v = jnp.ones((10,))

    matvec = build_hvp(f_crazy, x, linearise=False)
    matvec_lin = build_hvp(f_crazy, x, linearise=True)

    np.testing.assert_allclose(matvec(v), matvec_lin(v), atol=1e-8)



import warnings
from contextlib import contextmanager
from functools import wraps
from typing import Callable, Tuple, Dict, NamedTuple, Any, Optional, TypeVar, List, Union

import jax
from jax._src.typing import SupportsDType

__all__ = [
    'get_parameter',
    'get_state',
    'set_state',
    'transform_with_state',
    'transform',
    'convert_external_params',
    'wrap_random',
    'next_rng_key',
    'scope',
    'ScopedDict'
]


class ScopedDict:
    """
    prefixes all keys with a given scope {scope}.{key}
    """

    def __init__(self, _dict=None, _scopes=None):
        self.scopes: List[str] = _scopes or []
        self.dict = _dict or dict()

    def push_scope(self, scope):
        self.scopes.append(scope)

    def pop_scope(self):
        self.scopes.pop()

    @property
    def scope_prefix(self):
        return '.'.join(self.scopes)

    def to_dict(self):
        return self.dict

    def __repr__(self):
        return f"ScopedDict(scopes={repr(self.scopes)}, dict={repr(self.dict)})"

    def __getitem__(self, item):
        return self.dict[f"{self.scope_prefix}.{item}"]

    def __setitem__(self, key, value):
        self.dict[f"{self.scope_prefix}.{key}"] = value

    def __contains__(self, item):
        return f"{self.scope_prefix}.{item}" in self.dict

    def __iter__(self):
        return iter(self.dict)

    def __len__(self):
        return len(self.dict)

    def keys(self):
        return self.dict.keys()

    def values(self):
        return self.dict.values()

    def items(self):
        return self.dict.items()


# Add as pytree type

def scoped_dict_flatten(scoped_dict: ScopedDict):
    return (
        [
            scoped_dict.dict
        ],
        (scoped_dict.scopes,)
    )


def scoped_dict_unflatten(aux_data, children):
    [_dict] = children
    (_scopes,) = aux_data
    return ScopedDict(_dict=_dict, _scopes=_scopes)


jax.tree_util.register_pytree_node(
    ScopedDict,
    scoped_dict_flatten,
    scoped_dict_unflatten
)

MutableParams = Dict[str, jax.Array]
ImmutableParams = Dict[str, jax.Array]


class Ctx:
    def __init__(self, rng, stack):
        self.params = ScopedDict()
        self.states = ScopedDict()
        self._rng = rng
        self._stack = stack

    def next_rng_key(self):
        self._rng, new_rng = jax.random.split(self._rng)
        return new_rng

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._stack.remove(self)
        return

    def push_scope(self, scope):
        self.params.push_scope(scope)
        self.states.push_scope(scope)

    def pop_scope(self):
        self.params.pop_scope()
        self.states.pop_scope()


class GlobalContext:
    def __init__(self, rng: Optional[jax.Array] = None):
        self.stack: List[Ctx] = []

    def new(self, rng):
        new_ctx = Ctx(rng=rng, stack=self.stack)
        self.stack.append(new_ctx)
        return new_ctx

    @property
    def params(self):
        if len(self.stack) == 0:
            raise ValueError("No context available. Must use transform_with_state to create a context.")
        return self.stack[-1].params

    @property
    def states(self):
        if len(self.stack) == 0:
            raise ValueError("No context available. Must use transform_with_state to create a context.")
        return self.stack[-1].states

    def next_rng_key(self):
        if len(self.stack) == 0:
            raise ValueError("No context available. Must use transform_with_state to create a context.")
        return self.stack[-1].next_rng_key()

    def push_scope(self, scope):
        if len(self.stack) == 0:
            raise ValueError("No context available. Must use transform_with_state to create a context.")
        self.stack[-1].push_scope(scope)

    def pop_scope(self):
        if len(self.stack) == 0:
            raise ValueError("No context available. Must use transform_with_state to create a context.")
        self.stack[-1].pop_scope()


global_context = GlobalContext()

PT = TypeVar('PT')
InitType = Union[Callable[[Tuple[int, ...], SupportsDType], PT], Callable[[], PT]]


@contextmanager
def scope(name: str):
    """
    Create a new scope, to prefix parameters and states, as {current_scope}.{name}.{param_name}.

    Args:
        name: the name of the scope

    Returns:
        The scope
    """
    # Context manager
    global_context.push_scope(name)
    try:
        yield
    finally:
        global_context.pop_scope()


def default_init(shape: Tuple[int, ...], dtype: SupportsDType):
    raise NotImplementedError("No init provided.")


def get_parameter(name: str, shape: Optional[Tuple[int, ...]] = None, dtype: Optional[SupportsDType] = None, *,
                  init: InitType = default_init) -> PT:
    """
    Get a parameter variable.

    Args:
        name: the name of the parameter
        shape: the shape of the parameter must be provided if init is not a jax.Array
        dtype: the dtype of the parameter must be provided if init is not a jax.Array
        init: the initializer

    Returns:
        The parameter variable as a jax.Array=
    """
    if name not in global_context.params:
        if callable(init):

            if (shape is None) and (dtype is None):
                global_context.params[name] = init()
            else:
                global_context.params[name] = init(shape, dtype)
        else:
            warnings.warn(
                "Using a constant initializer for state. This is not recommended as it may induce closure issues.")
            global_context.params[name] = init

    return global_context.params[name]


ExtParam = TypeVar('ExtParam')


def convert_external_params(external_params: ExtParam, prefix: str) -> ExtParam:
    """
    Convert external parameters to context parameters. This can be used to convert haiku or flax parameters to
    jaxns parameters for using in models.

    Args:
        external_params: map of name -> value

    Returns:
        The context parameters
    """
    leaf_list, tree_def = jax.tree.flatten(external_params)

    def _unique_name(idx):
        return f"__{prefix}_{idx}"

    ctx_params = [get_parameter(_unique_name(idx), init=leaf) for idx, leaf in enumerate(leaf_list)]
    external_params_ctx = jax.tree_unflatten(tree_def, ctx_params)
    return external_params_ctx


def wrap_random(f):
    """
    Wrap a function to use a random number generator from the context.

    Args:
        f: the function to wrap

    Returns:
        The wrapped function
    """

    @wraps(f)
    def wrapped(*args, **kwargs):
        rng = next_rng_key()
        return f(rng, *args, **kwargs)

    return wrapped


def get_state(name: str, shape: Optional[Tuple[int, ...]] = None, dtype: Optional[SupportsDType] = None, *,
              init: InitType = default_init) -> PT:
    """
    Get a state variable.

    Args:
        name: the name of the state
        shape: the shape of the state must be provided if init is not a jax.Array
        dtype: the dtype of the state must be provided if init is not a jax.Array
        init: the initializer

    Returns:
        The state variable as a jax.Array
    """
    if name not in global_context.states:
        if callable(init):
            if (shape is None) and (dtype is None):
                global_context.states[name] = init()
            else:
                global_context.states[name] = init(shape, dtype)
        else:
            warnings.warn(
                "Using a constant initializer for state. This is not recommended as it may induce closure issues.")
            global_context.states[name] = init
    return global_context.states[name]


def set_state(name: str, value: PT):
    """
    Set a state variable.

    Args:
        name: the name of the state
        value: the value to set

    Returns:
        The state variable as a jax.Array
    """
    if name not in global_context.states:
        raise ValueError(f"State {name} not found.")
    # Ensure same pytree def
    tree_def = jax.tree.structure(global_context.states[name])
    value_tree_def = jax.tree.structure(value)
    if tree_def != value_tree_def:
        raise ValueError(f"Expected state with tree_def {tree_def} got {value_tree_def}.")
    global_context.states[name] = value


class InitWithStateReturn(NamedTuple):
    fn_val: Any
    params: MutableParams
    states: ImmutableParams


class ApplyWithStateReturn(NamedTuple):
    fn_val: Any
    states: ImmutableParams


class TransformedWithStateFn(NamedTuple):
    init: Callable[[jax.Array, ...], InitWithStateReturn]
    apply: Callable[[MutableParams, ImmutableParams, jax.Array, ...], ApplyWithStateReturn]


def transform_with_state(f: Callable) -> TransformedWithStateFn:
    """
    Transform a function to use parameters and states.

    Args:
        f: the function to transform

    Returns:
        A tuple of the init and apply functions
    """

    @wraps(f)
    def init(rng: jax.Array, *args, **kwargs) -> InitWithStateReturn:
        """
        Get initial parameters and states.

        Args:
            rng: the PRNGkey
            *args: args to the function
            **kwargs: kwargs to the function

        Returns:
            The output of the function at the given input, the parameters and the states
        """
        with global_context.new(rng) as ctx:
            return InitWithStateReturn(fn_val=f(*args, **kwargs), params=global_context.params,
                                       states=global_context.states)

    @wraps(f)
    def apply(params: MutableParams, states: ImmutableParams, rng: jax.Array, *args, **kwargs) -> ApplyWithStateReturn:
        """
        Apply the function with given parameters and states.

        Args:
            params: the parameters
            states: the states
            rng: the PRNGkey to apply at
            *args: args to function
            **kwargs: kwargs to function

        Returns:
            The output of the function at the given input and the states
        """
        with global_context.new(rng) as ctx:
            ctx.params = params
            ctx.states = states
            return ApplyWithStateReturn(fn_val=f(*args, **kwargs), states=ctx.states)

    return TransformedWithStateFn(init=init, apply=apply)


class InitReturn(NamedTuple):
    fn_val: Any
    params: MutableParams


class ApplyReturn(NamedTuple):
    fn_val: Any


class TransformedFn(NamedTuple):
    init: Callable[[jax.Array, ...], InitReturn]
    apply: Callable[[MutableParams, jax.Array, ...], ApplyReturn]


def transform(f: Callable) -> TransformedFn:
    """
    Transform a function to use parameters and states.

    Args:
        f: the function to transform

    Returns:
        A tuple of the init and apply functions
    """

    @wraps(f)
    def init(rng: jax.Array, *args, **kwargs) -> InitReturn:
        """
        Get initial parameters and states.

        Args:
            rng: the PRNGkey
            *args: args to the function
            **kwargs: kwargs to the function

        Returns:
            The output of the function at the given input, the parameters and the states
        """
        with global_context.new(rng) as ctx:
            return InitReturn(fn_val=f(*args, **kwargs), params=global_context.params)

    @wraps(f)
    def apply(params: MutableParams, rng: jax.Array, *args, **kwargs) -> ApplyReturn:
        """
        Apply the function with given parameters and states.

        Args:
            params: the parameters
            states: the states
            rng: the PRNGkey to apply at
            *args: args to function
            **kwargs: kwargs to function

        Returns:
            The output of the function at the given input and the states
        """
        with global_context.new(rng) as ctx:
            ctx.params = params
            return ApplyReturn(fn_val=f(*args, **kwargs))

    return TransformedFn(init=init, apply=apply)


def next_rng_key():
    """
    Get the next random number generator

    Returns:
        The next random number generator
    """
    return global_context.next_rng_key()



from typing import NamedTuple, Union, Any, Callable, Tuple, Dict, TypeVar

import jax
import numpy as np

__all__ = [
    'PRNGKey',
    'IntArray',
    'FloatArray',
    'BoolArray',
    'LikelihoodType',
    'UType',
    'XType',
    'LikelihoodInputType',
    'RandomVariableType',
    'MeasureType'
]

PRNGKey = jax.Array

Array = Union[
    jax.Array,  # JAX array type
    np.ndarray,  # NumPy array type
]
FloatArray = Union[
    jax.Array,  # JAX array type
    np.ndarray,  # NumPy array type
    float,  # valid scalars
]
IntArray = Union[
    jax.Array,  # JAX array type
    np.ndarray,  # NumPy array type
    int,  # valid scalars
]
BoolArray = Union[
    jax.Array,  # JAX array type
    np.ndarray,  # NumPy array type
    np.bool_, bool,  # valid scalars
]

Array.__doc__ = "Type annotation for JAX array-like objects, with no scalar types."

FloatArray.__doc__ = "Type annotation for JAX array-like objects, with float scalar types."

IntArray.__doc__ = "Type annotation for JAX array-like objects, with int scalar types."

BoolArray.__doc__ = "Type annotation for JAX array-like objects, with bool scalar types."

LikelihoodType = Callable[..., FloatArray]
RandomVariableType = TypeVar('RandomVariableType')
MeasureType = TypeVar('MeasureType')
LikelihoodInputType = Union[Tuple[Any, ...], Any]  # Likelihood conditional variables
UType = jax.Array  # Sample space type
WType = Tuple[jax.Array, ...]
XType = Dict[str, RandomVariableType]  # Prior variable type


class SignedLog(NamedTuple):
    """
    Represents a signed value in log-space
    """
    log_abs_val: jax.Array
    sign: Union[jax.Array, Any]


def isinstance_namedtuple(obj) -> bool:
    """
    Check if object is a namedtuple.

    Args:
        obj: object

    Returns:
        bool
    """
    return (
            isinstance(obj, tuple) and
            hasattr(obj, '_asdict') and
            hasattr(obj, '_fields')
    )



import inspect
import warnings
from typing import TypeVar, Callable, Optional, Tuple, List, Union, Any

import jax
import numpy as np
from jax import pmap, numpy as jnp, lax, NamedSharding
from jax._src.mesh import Mesh
from jax._src.partition_spec import PartitionSpec
from jax.experimental.mesh_utils import create_device_mesh

from jaxns.internals.mixed_precision import mp_policy


def replace_index(operand, update, start_index):
    """
    Replaces an index or slice with an update.
    If update is too big to respect start_index then start_index is shifted, which will give non-intuitive results.
    """
    if len(np.shape(operand)) != len(np.shape(update)):
        raise ValueError(
            f"Operand and update must have the same number of dimensions, got {len(np.shape(operand))} and {len(np.shape(update))}")
    start_index = jnp.asarray(start_index, mp_policy.index_dtype)
    start_indices = [start_index] + [jnp.asarray(0, start_index.dtype)] * (len(update.shape) - 1)
    return lax.dynamic_update_slice(operand, update.astype(operand.dtype), start_indices)


def get_index(operand, start_index, length):
    return lax.dynamic_slice(operand,
                             [start_index] + [jnp.asarray(0, mp_policy.index_dtype)] * (len(operand.shape) - 1),
                             (length,) + operand.shape[1:])


def prepare_func_args(f):
    """
    Takes a callable(a,b,...,z=Z) and prepares it into ``callable(**kwargs)``, such that only
    a,b,...,z are taken from ``**kwargs`` and the rest ignored.

    This allows ``f(**kwarg)`` to work even if ``f()`` is missing some keys from kwargs.

    Args:
        f: ``callable(a,b,...,z=Z)``

    Returns:
        ``callable(**kwargs)`` where ``**kwargs`` are the filtered for args of the original function.
    """
    if hasattr(f, "__old_name__"):
        raise ValueError(f"function {f.__old_name__} has already done prepare_func_args.")

    (args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults, annotations) = \
        inspect.getfullargspec(f)

    # TODO: this gets displayed each time we prepare a function. Using a cache would be cleaner for user.
    if varargs is not None:
        warnings.warn(f"Function {f.__name__} has *varargs parameter ({varargs}), and is being dropped.")
    if varkw is not None:
        warnings.warn(f"Function {f.__name__} has **varkw parameter ({varkw}), and is being dropped.")

    expected_keys = set(args + kwonlyargs)

    if defaults is None:  # no defaults
        num_defaults = 0
        defaults = ()
    else:
        num_defaults = len(defaults)

    num_non_default = len(args) - num_defaults

    def _f(**kwargs):
        # Replace any defaults with kwargs
        args_with_values = dict(zip(args[num_non_default:], defaults))
        if kwonlydefaults is not None:
            args_with_values.update(kwonlydefaults)
        args_with_values.update(kwargs)
        args_with_values = dict(filter(lambda item: item[0] in expected_keys, args_with_values.items()))
        for key in expected_keys:
            if key not in args_with_values:
                raise KeyError(f"{f.__name__} is missing argument {key}")

        return f(**args_with_values)

    _f.__doc__ = f.__doc__
    _f.__old_name__ = f.__name__
    return _f


F = TypeVar('F')
FV = TypeVar('FV')


def chunked_pmap(f: Callable[..., FV], chunk_size: Optional[int] = None, unroll: int = 1) -> Callable[..., FV]:
    """
    A version of pmap which chunks the input into smaller pieces to avoid memory issues.

    Args:
        f: callable
        chunk_size: the size of the chunks. Default is len(devices())
        unroll: the number of times to unroll the computation

    Returns:
        a chunked version of f
    """
    if chunk_size is None:
        chunk_size = len(jax.devices())

    def _f(*args, **kwargs):
        def queue(*args, **kwargs):
            """
            Distributes the computation in queues which are computed with scan.
            Args:
                *args:
            """

            def body(state, X):
                (args, kwargs) = X
                return state, f(*args, **kwargs)

            _, result = lax.scan(body, (), (args, kwargs), unroll=unroll)
            return result

        if chunk_size > 1:
            # Get from first leaf
            if len(args) > 0:
                batch_size = jax.tree.leaves(args)[0].shape[0]
            else:
                batch_size = jax.tree.leaves(kwargs)[0].shape[0]
            remainder = batch_size % chunk_size
            extra = (chunk_size - remainder) % chunk_size
            if extra > 0:
                (args, kwargs) = jax.tree.map(lambda x: _pad_extra(x, chunk_size), (args, kwargs))
            (args, kwargs) = jax.tree.map(
                lambda x: jnp.reshape(x, (chunk_size, x.shape[0] // chunk_size) + x.shape[1:]),
                (args, kwargs)
            )
            result = pmap(queue)(*args, **kwargs)  # [chunksize, batch_size // chunksize, ...]
            result = jax.tree.map(lambda x: jnp.reshape(x, (-1,) + x.shape[2:]), result)
            if extra > 0:
                result = jax.tree.map(lambda x: x[:-extra], result)
        else:
            result = queue(*args, **kwargs)
        return result

    _f.__doc__ = f.__doc__
    _f.__annotations__ = f.__annotations__
    return _f


def _pad_extra(arg, chunksize):
    N = arg.shape[0]
    remainder = N % chunksize
    if (remainder != 0) and (N > chunksize):
        # only pad if not a zero remainder
        extra = (chunksize - remainder) % chunksize
        arg = jnp.concatenate([arg] + [arg[0:1]] * extra, axis=0)
        N = N + extra
    else:
        extra = 0
    T = N // chunksize
    # arg = jnp.reshape(arg, (chunksize, N // chunksize) + arg.shape[1:])
    return arg


def prepad(a, chunksize: int):
    return jax.tree.map(lambda arg: _pad_extra(arg, chunksize), a)


T = TypeVar('T')


def remove_chunk_dim(py_tree: T) -> T:
    """
    Remove the chunk dimension from a pytree

    Args:
        py_tree: pytree to remove chunk dimension from

    Returns:
        pytree with chunk dimension removed
    """
    leaves = jax.tree.leaves(py_tree)

    # Check consistency
    for leaf in leaves:
        if len(leaf.shape) < 1:
            raise ValueError(f"Expected all leaves to have at least one dimension, got {leaf.shape}")
        if leaf.shape[0] != leaves[0].shape[0]:
            raise ValueError(
                f"Expected all leaves to have the same batch dimension, got {leaf.shape[0]} != {leaves[0].shape[0]}"
            )

    def _remove_chunk_dim(a):
        shape = list(a.shape)
        if len(shape) == 1:
            return a[0]
        shape = [shape[0] * shape[1]] + shape[2:]
        return jnp.reshape(a, shape)

    return jax.tree.map(_remove_chunk_dim, py_tree)


def add_chunk_dim(py_tree: T, chunk_size: int) -> T:
    """
    Add a chunk dimension to a pytree

    Args:
        py_tree: pytree to add chunk dimension to
        chunk_size: size of chunk dimension

    Returns:
        pytree with chunk dimension added
    """

    def _add_chunk_dim(a):
        shape = list(a.shape)
        shape = [chunk_size, shape[0] // chunk_size] + shape[1:]
        return jnp.reshape(a, shape)

    return jax.tree.map(_add_chunk_dim, py_tree)


def chunked_vmap(f, chunk_size: Optional[int] = None, unroll: int = 1):
    """
    A version of vmap which chunks the input into smaller pieces to avoid memory issues.

    Args:
        f: the function to be mapped
        chunk_size: the size of the chunks. Default is len(devices())
        unroll: the number of times to unroll the computation

    Returns:

    """
    if chunk_size is None:
        chunk_size = len(jax.devices())

    def _f(*args, **kwargs):
        def queue(*args, **kwargs):
            """
            Distributes the computation in queues which are computed with scan.
            Args:
                *args:
            """

            def body(state, X):
                (args, kwargs) = X
                return state, f(*args, **kwargs)

            _, result = lax.scan(f=body, init=(), xs=(args, kwargs), unroll=unroll)
            return result

        if chunk_size > 1:
            # Get from first leaf
            if len(args) > 0:
                batch_size = jax.tree.leaves(args)[0].shape[0]
            else:
                batch_size = jax.tree.leaves(kwargs)[0].shape[0]
            remainder = batch_size % chunk_size
            extra = (chunk_size - remainder) % chunk_size
            if extra > 0:
                (args, kwargs) = jax.tree.map(lambda x: _pad_extra(x, chunk_size), (args, kwargs))
            (args, kwargs) = jax.tree.map(
                lambda x: jnp.reshape(x, (chunk_size, x.shape[0] // chunk_size) + x.shape[1:]),
                (args, kwargs)
            )
            result = jax.vmap(queue)(*args, **kwargs)  # [chunksize, batch_size // chunksize, ...]
            result = jax.tree.map(lambda x: jnp.reshape(x, (-1,) + x.shape[2:]), result)
            if extra > 0:
                result = jax.tree.map(lambda x: x[:-extra], result)
        else:
            result = queue(*args, **kwargs)
        return result

    _f.__doc__ = f.__doc__
    _f.__annotations__ = f.__annotations__
    return _f


PT = TypeVar('PT')


def pytree_unravel(example_tree: PT) -> Tuple[Callable[[PT], jax.Array], Callable[[jax.Array], PT]]:
    """
    Returns functions to ravel and unravel a pytree.

    Args:
        example_tree: a pytree to be unravelled

    Returns:
        ravel_fun: a function to ravel the pytree
        unravel_fun: a function to unravel
    """
    leaf_list, tree_def = jax.tree.flatten(example_tree)

    sizes = [np.size(leaf) for leaf in leaf_list]
    shapes = [np.shape(leaf) for leaf in leaf_list]
    dtypes = [leaf.dtype for leaf in leaf_list]

    def ravel_fun(pytree: PT) -> jax.Array:
        leaf_list, tree_def = jax.tree.flatten(pytree)
        # promote types to common one
        common_dtype = jnp.result_type(*dtypes)
        leaf_list = [leaf.astype(common_dtype) for leaf in leaf_list]
        return jnp.concatenate([lax.reshape(leaf, (size,)) for leaf, size in zip(leaf_list, sizes)])

    def unravel_fun(flat_array: jax.Array) -> PT:
        leaf_list = []
        start = 0
        for size, shape, dtype in zip(sizes, shapes, dtypes):
            leaf_list.append(lax.reshape(flat_array[start:start + size], shape).astype(dtype))
            start += size
        return jax.tree.unflatten(tree_def, leaf_list)

    return ravel_fun, unravel_fun


def pytree_unpack(example_tree: PT) -> Tuple[Callable[[PT], List[jax.Array]], Callable[[List[jax.Array]], PT]]:
    """
    Returns functions to ravel and unravel a pytree.
    """
    leaf_list, tree_def = jax.tree.flatten(example_tree)

    def pack_fun(pytree: PT) -> List[jax.Array]:
        leaf_list, tree_def = jax.tree.flatten(pytree)
        return leaf_list

    def unpack_fun(leaf_list: List[jax.Array]) -> PT:
        return jax.tree.unflatten(tree_def, leaf_list)

    return pack_fun, unpack_fun


PV = TypeVar('PV')


class PyTree:
    """
    For acting on W space.
    """

    def __init__(self, tree: PV):
        self.tree = tree

    def __add__(self, other: PV) -> PV:
        return jax.tree.map(lambda x, y: x + y, self.tree, other)

    def __sub__(self, other: PV) -> PV:
        return jax.tree.map(lambda x, y: x - y, self.tree, other)

    def __mul__(self, other: PV) -> PV:
        return jax.tree.map(lambda x, y: x * y, self.tree, other)

    def __truediv__(self, other: PV) -> PV:
        return jax.tree.map(lambda x, y: x / y, self.tree, other)

    def __pow__(self, other: PV) -> PV:
        return jax.tree.map(lambda x, y: x ** y, self.tree, other)

    def __neg__(self):
        return jax.tree.map(lambda x: -x, self.tree)


def create_mesh(shape, axis_names, devices=None):
    """
    Create a mesh from a shape and axis names.

    Args:
        shape: the shape of the mesh, total size must evenly divide number of devices.
        axis_names: the axis names of the mesh.
        devices: the devices to use, if None, uses all devices.

    Returns:
        the mesh
    """
    if len(shape) != len(axis_names):
        raise ValueError(f"Shape {shape} and axis names {axis_names} must have the same length.")
    mesh_size = int(np.prod(shape))
    if devices is None:
        devices = jax.devices()
        if mesh_size < len(devices):
            devices = devices[:mesh_size]
    if mesh_size % len(devices) != 0:
        raise ValueError(f"Mesh size {mesh_size} must evenly divide number of devices {len(devices)}.")
    mesh_devices = create_device_mesh(mesh_shape=shape, devices=devices)
    mesh = Mesh(mesh_devices, axis_names=axis_names)
    return mesh


SPT = TypeVar('SPT')


def tree_device_put(tree: SPT, mesh: Mesh, axis_names: Tuple[Union[str, None], ...]) -> SPT:
    """
    Put a pytree on a device.

    Args:
        tree: the pytree to put on a device.
        mesh: the mesh to put the pytree on.
        axis_names: the axis names of the mesh.

    Returns:
        the pytree on the device.
    """
    sharding = NamedSharding(mesh, PartitionSpec(*axis_names))
    return jax.tree.map(lambda x: jax.device_put(x, sharding), tree)


BUX = TypeVar('BUX', bound=Union[jax.Array, Any])


def block_until_ready(x: BUX) -> BUX:
    return jax.block_until_ready(x)



from typing import NamedTuple, TypeVar, Tuple, Callable, Any

import jax
import jax.numpy as jnp

from jaxns.experimental.solvers.ad_utils import tree_neg, tree_vdot_real_part, tree_scalar_mul, tree_add, tree_sub, \
    build_hvp
from jaxns.experimental.solvers.cg import cg_solve
from jaxns.internals.types import IntArray, FloatArray, BoolArray

# ----------------------------------------------------------------
# Type helpers
# ----------------------------------------------------------------
DomainType = TypeVar("DomainType")  # parameter pytree
ObjectiveRet = TypeVar("ObjectiveRet")  # scalar objective (FloatArray or 0-D ndarray)

CT = TypeVar("CT")
_CT = TypeVar("_CT")


# ----------------------------------------------------------------
# Utility: split complex pytrees into real pairs so Wirtinger calculus
#          works out of the box with JAX’s real autodiff.
# ----------------------------------------------------------------
def convert_to_real(x: CT) -> Tuple[_CT, Callable[[_CT], CT]]:
    """Return a real-valued twin of `x`   and   a merge-back function."""

    def _maybe_split(a: jax.Array | Any):
        if isinstance(a, jax.Array) and jnp.iscomplexobj(a):
            return (a.real, a.imag)
        return a

    leaves, treedef = jax.tree.flatten(x)
    split_leaves = jax.tree.map(_maybe_split, leaves)

    def merge(split_x: _CT) -> CT:
        def _maybe_merge(a):
            if isinstance(a, tuple):
                return jax.lax.complex(a[0], a[1])
            return a

        merged = list(map(_maybe_merge, split_x))
        return jax.tree.unflatten(treedef, merged)

    return split_leaves, merge


# ----------------------------------------------------------------
# Diagnostics – patterned after LMDiagnostic
# ----------------------------------------------------------------
class NewtonDiagnostic(NamedTuple):
    iteration: IntArray
    g_norm: FloatArray  # |∇f|
    mu: FloatArray  # damping parameter
    damping: FloatArray  # g_norm / mu
    cg_iters: IntArray
    f: FloatArray  # f(x_k)
    f_prop: FloatArray  # f(x_k + δx_k)
    f_quad: FloatArray  # quadratic model at proposal
    delta_f_pred: FloatArray  # predicted decrease
    delta_f_actual: FloatArray  # actual decrease
    gain_ratio: FloatArray  # delta_f_actual / delta_f_pred
    accepted: BoolArray
    in_trust_region: BoolArray
    delta_x_norm: FloatArray  # ‖δx‖
    ddelta_x_norm: FloatArray  # ‖δx – δx⁰‖


# ----------------------------------------------------------------
# Main solver
# ----------------------------------------------------------------
def newton_cg_solver(
        obj_fn: Callable[..., ObjectiveRet],
        x0: DomainType,
        args: tuple = (),
        maxiter: int = 100,
        maxiter_cg: int = 100,
        gtol: float = 3e-5,
        p_accept: float = 0.01,
        p_lower: float = 0.25,
        p_upper: float = 1.10,
        mu_init: float = 1.0,
        mu_min: float = 1e-6,
        mu_in_factor: float = 5,
        mu_out_factor: float = 0.1,
        approx_hvp: bool = False,  # reuse H·v between rejections
        verbose: bool = False,
) -> Tuple[DomainType, NewtonDiagnostic]:
    """
    Trust-region Newton-CG minimiser.

    Identical call signature and adaptive-μ logic as `lm_solver`, but uses a
    scalar objective instead of residuals, and solves
        (H  +  damping·I) δx = -∇f
    by CG with Hessian–vector products.

    Returns
    -------
    x_final : pytree matching `x0`     (merged back to complex if needed)
    diagnostics : NewtonDiagnostic[...] array with length = `maxiter`
    """
    # ---- 1.  Handle complex inputs -----------------------------------------
    x0_real, merge_back = convert_to_real(x0)

    # Wrap obj_fn so that it consumes / produces purely real pytrees
    def _obj_fn(x):
        val = obj_fn(merge_back(x), *args)
        if not isinstance(val, jax.Array):
            raise RuntimeError(f"Objective function must return a JAX scalar array, got {type(val)}.")
        if jnp.ndim(val) != 0:
            raise RuntimeError("Objective function must return a scalar.")
        return val

    # ---- 2.  State container -----------------------------------------------
    class NState(NamedTuple):
        x: DomainType  # current parameters (real)
        f: FloatArray  # f(x)
        g: DomainType  # -∇f  (descent direction)
        g_norm: FloatArray
        mu: FloatArray
        delta_x_prev: DomainType  # δx⁻¹
        delta_x_prev2: DomainType  # δx⁻²
        iter: IntArray

    # ---- 3.  Helpers --------------------------------------------------------
    def _gradient(x):
        return jax.grad(_obj_fn)(x)

    def _initial_state(x):
        f0 = _obj_fn(x)
        grad_f = _gradient(x)
        g0 = tree_neg(grad_f)  # -∇f
        g_norm0 = jnp.sqrt(tree_vdot_real_part(g0, g0))
        g_unit = tree_scalar_mul(1.0 / (g_norm0 + 1e-12), g0)

        # Simple backtracking line search along -∇f to pick starting μ
        def ls_cond(mu):
            step = tree_scalar_mul(mu, g_unit)
            f_new = _obj_fn(tree_add(x, step))
            return (f_new >= f0) & (mu > mu_min)

        def ls_body(mu):
            return mu * mu_out_factor

        mu0 = jax.lax.while_loop(ls_cond, ls_body, mu_init)

        z = jax.tree.map(jnp.zeros_like, x)
        return NState(
            x=x, f=f0, g=g0, g_norm=g_norm0, mu=mu0,
            delta_x_prev=z, delta_x_prev2=z, iter=0
        )

    # ---- 4.  Iteration ------------------------------------------------------
    def cond_fn(carry):
        state, _ = carry
        return (state.g_norm > gtol) & (state.iter < maxiter)

    def step_fn(state: NState):
        # 4.1  Warm-start: δx⁰ = 2δx⁻¹ − δx⁻²
        delta_x0 = tree_sub(tree_add(state.delta_x_prev, state.delta_x_prev),
                            state.delta_x_prev2)

        # 4.2  Build Hessian-vector product operator (maybe reused)
        hvp = build_hvp(_obj_fn, state.x, linearise=True)
        damping = state.g_norm / state.mu

        def A_op(v):
            return tree_add(hvp(v), tree_scalar_mul(damping, v))

        # 4.3  Solve Newton system with CG
        delta_x, cg_diag = cg_solve(
            A=A_op, b=state.g, x0=delta_x0,
            maxiter=maxiter_cg, tol=1e-5, atol=0.0
        )

        # 4.4  Book-keeping norms
        delta_x_norm = jnp.sqrt(tree_vdot_real_part(delta_x, delta_x))
        ddelta_x = tree_sub(delta_x, delta_x0)
        ddelta_x_norm = jnp.sqrt(tree_vdot_real_part(ddelta_x, ddelta_x))

        # 4.5  Evaluate objective at proposal
        x_prop = tree_add(state.x, delta_x)
        f_prop = _obj_fn(x_prop)

        # Quadratic model prediction
        hvp_dx = hvp(delta_x)
        quad_term = 0.5 * tree_vdot_real_part(delta_x, hvp_dx)
        lin_term = tree_vdot_real_part(state.g, delta_x)  # g·δx
        f_quad = state.f - (lin_term - quad_term)  # f(x)+...
        delta_f_pred = state.f - f_quad  # should be >0
        delta_f_actual = state.f - f_prop
        gain_ratio = delta_f_actual / delta_f_pred
        gain_ratio = jnp.where(jnp.isnan(gain_ratio), 0.0, gain_ratio)  # handle NaN

        # 4.6  Trust-region logic
        in_trust = (delta_f_pred > 0) & (delta_f_actual > p_lower * delta_f_pred) & (
                delta_f_actual < p_upper * delta_f_pred)
        new_mu = jax.lax.select(in_trust, mu_in_factor * state.mu, state.mu * mu_out_factor)
        new_mu = jnp.maximum(new_mu, mu_min)

        accepted = (delta_f_pred > 0) & (delta_f_actual > p_accept * delta_f_pred)

        # 4.7  Diagnostics -----------------------------------------------------
        diag = NewtonDiagnostic(
            iteration=state.iter,
            g_norm=state.g_norm,
            mu=state.mu,
            damping=damping,
            cg_iters=cg_diag.iterations,
            f=state.f,
            f_prop=f_prop,
            f_quad=f_quad,
            delta_f_pred=delta_f_pred,
            delta_f_actual=delta_f_actual,
            gain_ratio=gain_ratio,
            accepted=accepted,
            in_trust_region=in_trust,
            delta_x_norm=delta_x_norm,
            ddelta_x_norm=ddelta_x_norm
        )
        if verbose:
            jax.debug.print(
                "iter {iteration:3d}  f={f:9.3e}  |g|={g_norm:9.3e}  "
                "μ={mu:9.3e}  r={gain_ratio:6.2f}  "
                "Δf_pred={delta_f_pred:9.3e}  Δf_act={delta_f_actual:9.3e}  "
                "CG={cg_iters}",
                iteration=state.iter,
                f=state.f, g_norm=state.g_norm, mu=state.mu,
                gain_ratio=diag.gain_ratio,
                delta_f_pred=diag.delta_f_pred,
                delta_f_actual=diag.delta_f_actual,
                cg_iters=diag.cg_iters
            )

        # 4.8  Accept / reject step
        (
            x_new, f_new, delta_x_prev_new, delta_x_prev2_new
        ) = jax.tree.map(
            lambda a, b: jax.lax.select(accepted, a, b),
            (x_prop, f_prop, delta_x, state.delta_x_prev),
            (state.x, state.f, state.delta_x_prev, state.delta_x_prev2)
        )

        # 4.9  Gradient at (possibly) new point
        if approx_hvp:
            # keep previous hvp & gradient
            g_new = state.g
            g_norm_new = state.g_norm
        else:
            grad_f_new = _gradient(x_new)
            g_new = tree_neg(grad_f_new)
            g_norm_new = jnp.sqrt(tree_vdot_real_part(g_new, g_new))

        # 4.10  Next state
        return NState(
            x=x_new, f=f_new, g=g_new, g_norm=g_norm_new, mu=new_mu,
            delta_x_prev=delta_x_prev_new, delta_x_prev2=delta_x_prev2_new,
            iter=state.iter + 1
        ), diag

    # --------------------------------------------------------------------
    # 5.  Main while-loop with diagnostic accumulation
    # --------------------------------------------------------------------
    init_state = _initial_state(x0_real)

    # prototype diag for shape inference
    diag_aval = jax.eval_shape(lambda s: step_fn(s)[1], init_state)
    empty_diag = jax.tree.map(
        lambda arr: jnp.zeros((maxiter,) + arr.shape, arr.dtype),
        diag_aval
    )

    def body_fn(carry):
        st, diag_arr = carry
        new_st, new_d = step_fn(st)
        diag_arr = jax.tree.map(
            lambda arr, d: arr.at[st.iter].set(d),
            diag_arr, new_d
        )
        return new_st, diag_arr

    final_state, final_diag = jax.lax.while_loop(
        cond_fn, body_fn, (init_state, empty_diag)
    )

    # Merge complex components back to the user space
    return merge_back(final_state.x), final_diag



import warnings
from typing import Union, Tuple

from jax import numpy as jnp

from jaxns.internals.mixed_precision import mp_policy
from jaxns.internals.stats import linear_to_log_stats, effective_sample_size_kish
from jaxns.internals.types import BoolArray, IntArray
from jaxns.nested_samplers.common.types import TerminationConditionDisjunction, TerminationConditionConjunction, \
    TerminationRegister, TerminationCondition


def determine_termination(
        term_cond: Union[TerminationConditionDisjunction, TerminationConditionConjunction, TerminationCondition],
        termination_register: TerminationRegister) -> Tuple[BoolArray, IntArray]:
    """
    Determine if termination should happen. Termination Flags are bits:
        0-bit -> 1: used maximum allowed number of samples
        1-bit -> 2: evidence uncert below threshold
        2-bit -> 4: live points evidence below threshold
        3-bit -> 8: effective sample size big enough
        4-bit -> 16: used maxmimum allowed number of likelihood evaluations
        5-bit -> 32: maximum log-likelihood contour reached
        6-bit -> 64: sampler efficiency too low
        7-bit -> 128: entire live-points set is a single plateau
        8-bit -> 256: relative spread of live points < rtol
        9-bit -> 512: absolute spread of live points < atol
        10-bit -> 1024: no seed points left
        11-bit -> 2048: XL < max(XL) * peak_XL_frac

    Multiple flags are summed together

    Args:
        term_cond: termination condition
        termination_register: register of termination variables to check against termination condition

    Returns:
        boolean done signal, and termination reason
    """

    termination_reason = jnp.asarray(0, mp_policy.count_dtype)
    done = jnp.asarray(False, jnp.bool_)

    def _set_done_bit(bit_done, bit_reason, done, termination_reason):
        if bit_done.size > 1:
            raise RuntimeError("bit_done must be a scalar.")
        done = jnp.bitwise_or(bit_done, done)
        termination_reason += jnp.where(bit_done,
                                        jnp.asarray(2 ** bit_reason, mp_policy.count_dtype),
                                        jnp.asarray(0, mp_policy.count_dtype))
        return done, termination_reason

    if isinstance(term_cond, TerminationConditionConjunction):
        for c in term_cond.conds:
            _done, _reason = determine_termination(term_cond=c, termination_register=termination_register)
            done = jnp.bitwise_and(_done, done)
            termination_reason = jnp.bitwise_and(_reason, termination_reason)
        return done, termination_reason

    if isinstance(term_cond, TerminationConditionDisjunction):
        for c in term_cond.conds:
            _done, _reason = determine_termination(term_cond=c, termination_register=termination_register)
            done = jnp.bitwise_or(_done, done)
            termination_reason = jnp.bitwise_or(_reason, termination_reason)
        return done, termination_reason

    if term_cond.live_evidence_frac is not None:
        warnings.warn("live_evidence_frac is deprecated, use dlogZ instead.")

    if term_cond.max_samples is not None:
        # used all points
        reached_max_samples = termination_register.num_samples_used >= term_cond.max_samples
        done, termination_reason = _set_done_bit(reached_max_samples, 0,
                                                 done=done, termination_reason=termination_reason)

    if term_cond.evidence_uncert is not None:
        _, log_Z_var = linear_to_log_stats(
            log_f_mean=termination_register.evidence_calc_with_remaining.log_Z_mean,
            log_f2_mean=termination_register.evidence_calc_with_remaining.log_Z2_mean)
        evidence_uncert_low_enough = log_Z_var <= jnp.square(term_cond.evidence_uncert)
        done, termination_reason = _set_done_bit(evidence_uncert_low_enough, 1,
                                                 done=done, termination_reason=termination_reason)

    if term_cond.dlogZ is not None:
        # (Z_remaining + Z_current) / Z_remaining < exp(dlogZ)
        log_Z_mean_1, log_Z_var_1 = linear_to_log_stats(
            log_f_mean=termination_register.evidence_calc_with_remaining.log_Z_mean,
            log_f2_mean=termination_register.evidence_calc_with_remaining.log_Z2_mean)

        log_Z_mean_0, log_Z_var_0 = linear_to_log_stats(
            log_f_mean=termination_register.evidence_calc.log_Z_mean,
            log_f2_mean=termination_register.evidence_calc.log_Z2_mean)

        small_remaining_evidence = jnp.less(
            log_Z_mean_1 - log_Z_mean_0, term_cond.dlogZ
        )
        done, termination_reason = _set_done_bit(small_remaining_evidence, 2,
                                                 done=done, termination_reason=termination_reason)

    if term_cond.ess is not None:
        # Kish's ESS = [sum weights]^2 / [sum weights^2]
        ess = effective_sample_size_kish(termination_register.evidence_calc_with_remaining.log_Z_mean,
                                         termination_register.evidence_calc_with_remaining.log_dZ2_mean)
        ess_reached = ess >= term_cond.ess
        done, termination_reason = _set_done_bit(ess_reached, 3,
                                                 done=done, termination_reason=termination_reason)

    if term_cond.max_num_likelihood_evaluations is not None:
        num_likelihood_evaluations = jnp.sum(termination_register.num_likelihood_evaluations)
        too_max_likelihood_evaluations = num_likelihood_evaluations >= term_cond.max_num_likelihood_evaluations
        done, termination_reason = _set_done_bit(too_max_likelihood_evaluations, 4,
                                                 done=done, termination_reason=termination_reason)

    if term_cond.log_L_contour is not None:
        likelihood_contour_reached = termination_register.log_L_contour >= term_cond.log_L_contour
        done, termination_reason = _set_done_bit(likelihood_contour_reached, 5,
                                                 done=done, termination_reason=termination_reason)

    if term_cond.efficiency_threshold is not None:
        efficiency_too_low = termination_register.efficiency < term_cond.efficiency_threshold
        done, termination_reason = _set_done_bit(efficiency_too_low, 6,
                                                 done=done, termination_reason=termination_reason)

    done, termination_reason = _set_done_bit(termination_register.plateau, 7,
                                             done=done, termination_reason=termination_reason)

    if term_cond.rtol is not None:
        relative_spread_low = termination_register.relative_spread < term_cond.rtol
        done, termination_reason = _set_done_bit(relative_spread_low, 8,
                                                 done=done, termination_reason=termination_reason)

    if term_cond.atol is not None:
        absolute_spread_low = termination_register.absolute_spread < term_cond.atol
        done, termination_reason = _set_done_bit(absolute_spread_low, 9,
                                                 done=done, termination_reason=termination_reason)

    done, termination_reason = _set_done_bit(termination_register.no_seed_points, 10,
                                             done=done, termination_reason=termination_reason)

    if term_cond.peak_XL_frac is not None:
        log_XL = termination_register.evidence_calc.log_X_mean + termination_register.evidence_calc.log_L
        peak_log_XL = termination_register.peak_log_XL
        XL_reduction_reached = log_XL < peak_log_XL + jnp.log(term_cond.peak_XL_frac)
        done, termination_reason = _set_done_bit(XL_reduction_reached, 11,
                                                 done=done, termination_reason=termination_reason)

    return done, termination_reason



import dataclasses
from typing import Optional, Tuple, Union, List

import jax
import jax.numpy as jnp
import numpy as np
import tensorflow_probability.substrates.jax as tfp
from jax import core
from jaxlib import xla_client

from jaxns.framework.bases import BaseAbstractModel
from jaxns.internals.logging import logger
from jaxns.internals.mixed_precision import mp_policy
from jaxns.internals.types import PRNGKey, IntArray
from jaxns.nested_samplers.abc import AbstractNestedSampler
from jaxns.nested_samplers.common.types import TerminationCondition, NestedSamplerResults, \
    NestedSamplerState
from jaxns.nested_samplers.sharded import ShardedStaticNestedSampler
from jaxns.plotting import plot_cornerplot, plot_diagnostics
from jaxns.samplers.uni_slice_sampler import UniDimSliceSampler
from jaxns.utils import summary, save_results, load_results

tfpd = tfp.distributions

__all__ = [
    'NestedSampler'
]


@dataclasses.dataclass(eq=False)
class NestedSampler:
    """
    A static nested sampler that uses 1-dimensional slice sampler for the sampling step.
    Uses the phantom-powered algorithm. A robust default choice is provided for all parameters.

    s,k,c are defined in the paper: https://arxiv.org/abs/2312.11330

    Args:
        model: a model to perform nested sampling on
        max_samples: maximum number of samples to take
        num_live_points: approximate number of live points to use. Defaults is c * (k + 1).
        s: number of slices to use per dimension. Defaults to 4.
        k: number of phantom samples to use. Defaults to 0.
        c: number of parallel Markov-chains to use. Defaults to 20 * D.
        devices: devices to use. Defaults to all available devices.
        difficult_model: if True, uses more robust default settings. Defaults to False.
        parameter_estimation: if True, uses more robust default settings for parameter estimation. Defaults to False.
        shell_fraction: fraction of the shell to use for the slice sampler. Defaults to 0.5.
        gradient_guided: if True, uses gradient guided sampling. Defaults to False.
        init_efficiency_threshold: if > 0 then use uniform sampling first down to this acceptance efficiency.
            0 turns it off.
        verbose: whether to log progress.
    """
    model: BaseAbstractModel
    max_samples: Optional[Union[int, float]] = None
    num_live_points: Optional[int] = None
    num_slices: Optional[int] = None
    s: Optional[Union[int, float]] = None
    k: Optional[int] = None
    c: Optional[int] = None
    devices: Optional[List[xla_client.Device]] = None
    difficult_model: bool = False
    parameter_estimation: bool = False
    shell_fraction: float = 0.5
    gradient_guided: bool = False
    init_efficiency_threshold: float = 0.1
    verbose: bool = False

    def __post_init__(self):
        # Determine number of slices per acceptance
        if self.num_slices is None:
            if self.difficult_model:
                self.s = 10 if self.s is None else float(self.s)
            else:
                self.s = 5 if self.s is None else float(self.s)
            if self.s <= 0:
                raise ValueError(f"Expected s > 0, got s={self.s}")
            self.num_slices = self.model.U_ndims * self.s
        self.num_slices = int(self.num_slices)

        # Determine number of phantom samples
        if self.parameter_estimation:
            max_k = self.s * self.model.U_ndims - 1
            self.k = min(self.model.U_ndims, max_k) if self.k is None else int(self.k)
        else:
            self.k = 0 if self.k is None else int(self.k)
        if not (0 <= self.k < self.num_slices):
            raise ValueError(
                f"Expected 0 <= k < num_slices, got k={self.k}, num_slices={self.num_slices}, U_ndims={self.model.U_ndims}")

        # Determine number of parallel Markov-chains
        if self.num_live_points is not None:
            self.c = max(1, int(self.num_live_points / (self.k + 1)))
            logger.info(f"Number of Markov-chains set to: {self.c}")
        else:
            if self.difficult_model:
                self.c = 100 * self.model.U_ndims if self.c is None else int(self.c)
            else:
                self.c = 30 * self.model.U_ndims if self.c is None else int(self.c)
            if self.c <= 0:
                raise ValueError(f"Expected c > 0, got c={self.c}")

        # Sanity check for max_samples (should be able to at least do one shrinkage)
        if self.max_samples is None:
            # Default to 100 shrinkages.
            self.max_samples = self.c * (self.k + 1) * 100
        self.max_samples = int(self.max_samples)

        self._nested_sampler = ShardedStaticNestedSampler(
            model=self.model,
            num_live_points=self.c,
            max_samples=self.max_samples,
            sampler=UniDimSliceSampler(
                model=self.model,
                num_slices=self.num_slices,
                num_phantom_save=self.k,
                midpoint_shrink=not self.difficult_model,
                gradient_guided=self.gradient_guided,
                perfect=True
            ),
            init_efficiency_threshold=self.init_efficiency_threshold,
            shell_fraction=self.shell_fraction,
            devices=self.devices,
            verbose=self.verbose,

        )
        # Back propagate any updates here
        self.num_live_points = self._nested_sampler.num_live_points

        # Post-analysis utilities
        self.summary = summary
        self.plot_cornerplot = plot_cornerplot
        self.plot_diagnostics = plot_diagnostics
        self.save_results = save_results
        self.load_results = load_results

    @property
    def nested_sampler(self) -> AbstractNestedSampler:
        return self._nested_sampler

    def __call__(self, key: PRNGKey, term_cond: Optional[TerminationCondition] = None) -> Tuple[
        IntArray, NestedSamplerState]:
        """
        Performs nested sampling with the given termination conditions.

        Args:
            key: PRNGKey
            term_cond: termination conditions. If not given, see `TerminationCondition` for defaults.

        Returns:
            termination reason, state
        """
        if term_cond is None:
            if self.parameter_estimation:
                term_cond = TerminationCondition(
                    peak_XL_frac=jnp.asarray(0.1, mp_policy.measure_dtype),
                    max_samples=jnp.asarray(jnp.iinfo(mp_policy.count_dtype).max, mp_policy.count_dtype)
                )
            else:
                term_cond = TerminationCondition(
                    dlogZ=jnp.asarray(np.log(1. + 1e-3), mp_policy.measure_dtype),
                    max_samples=jnp.asarray(jnp.iinfo(mp_policy.count_dtype).max, mp_policy.count_dtype)
                )
        term_cond = term_cond._replace(
            max_samples=(
                jnp.minimum(term_cond.max_samples, jnp.asarray(self._nested_sampler.max_samples, mp_policy.count_dtype))
                if term_cond.max_samples is not None
                else jnp.asarray(self._nested_sampler.max_samples, mp_policy.count_dtype)
            )
        )
        termination_reason, termination_register, state = self._nested_sampler._run(
            key=key,
            term_cond=term_cond
        )
        return termination_reason, state

    def to_results(self, termination_reason: IntArray, state: NestedSamplerState,
                   trim: bool = True) -> NestedSamplerResults:
        """
        Convert the state to results.

        Note: Requires static context.

        Args:
            termination_reason: termination reason
            state: state to convert
            trim: if True, trims the results to the number of samples taken, requires static context.

        Returns:
            results
        """
        return self._nested_sampler._to_results(
            termination_reason=termination_reason,
            state=state,
            trim=trim
        )

    @staticmethod
    def trim_results(results: NestedSamplerResults) -> NestedSamplerResults:
        """
        Trims the results to the number of samples taken. Requires static context.

        Args:
            results: results to trim

        Returns:
            trimmed results
        """

        if isinstance(results.total_num_samples, core.Tracer):
            raise RuntimeError("Tracer detected, but expected imperative context.")

        def trim(x):
            if x.size > 1:
                return x[:results.total_num_samples]
            return x

        results = jax.tree.map(trim, results)
        return results


DefaultNestedSampler = NestedSampler



from abc import abstractmethod
from typing import NamedTuple, Tuple, TypeVar, Generic

from jax import random

from jaxns.internals.types import FloatArray
from jaxns.internals.types import PRNGKey
from jaxns.nested_samplers.common.types import Sample
from jaxns.samplers.abc import AbstractSampler


class SeedPoint(NamedTuple):
    U0: FloatArray
    log_L0: FloatArray


T = TypeVar('T')


class BaseAbstractRejectionSampler(AbstractSampler[T], Generic[T]):
    ...


class BaseAbstractMarkovSampler(AbstractSampler[T], Generic[T]):
    """
    A sampler that conditions off a known satisfying point, e.g. a seed point.
    """

    @abstractmethod
    def get_sample_from_seed(self, key: PRNGKey, seed_point: SeedPoint, log_L_constraint: FloatArray,
                             sampler_state: T) -> Tuple[Sample, Sample]:
        """
        Produce a single i.i.d. sample from the model within the log_L_constraint.

        Args:
            key: PRNGkey
            seed_point: function that gets the next sample from a seed point
            log_L_constraint: the constraint to sample within
            sampler_state: the data pytree needed and produced by the sampler

        Returns:
            an i.i.d. sample, and batched phantom samples
        """
        ...

    @abstractmethod
    def get_seed_point(self, key: PRNGKey, sampler_state: T,
                       log_L_constraint: FloatArray) -> SeedPoint:
        """
        Samples a seed point from the live points.

        Args:
            key: PRNGKey
            sampler_state: the current sampler state
            log_L_constraint: a log-L constraint to sample within. Must always be at least one sample in front above
                this to avoid infinite loop.

        Returns:
            a seed point
        """
        ...

    def _get_sample(self, key: PRNGKey, log_L_constraint: FloatArray, sampler_state: T) -> Tuple[Sample, Sample]:
        sample_key, seed_key = random.split(key, 2)
        seed_point = self.get_seed_point(
            key=seed_key,
            sampler_state=sampler_state,
            log_L_constraint=log_L_constraint
        )
        return self.get_sample_from_seed(
            key=sample_key,
            seed_point=seed_point,
            log_L_constraint=log_L_constraint,
            sampler_state=sampler_state
        )



from functools import partial

import jax
import jax.numpy as jnp

_dot = partial(jnp.dot, precision=jax.lax.Precision.HIGHEST)
_vdot = partial(jnp.vdot, precision=jax.lax.Precision.HIGHEST)
_einsum = partial(jnp.einsum, precision=jax.lax.Precision.HIGHEST)


def tree_dot(x, y):
    dots = jax.tree.leaves(jax.tree.map(jnp.vdot, x, y))
    return sum(dots[1:], start=dots[0])


def tree_norm(x):
    return jnp.sqrt(tree_dot(x, x).real)


def tree_mul(x, y):
    return jax.tree.map(jax.lax.mul, x, y)


def tree_div(x, y):
    return jax.tree.map(jax.lax.div, x, y)


# aliases for working with pytrees
def _vdot_real_part(x, y):
    """Vector dot-product guaranteed to have a real valued result despite
       possibly complex input. Thus neglects the real-imaginary cross-terms.
       The result is a real float.
    """
    # all our uses of vdot() in CG are for computing an operator of the form
    #  z^H M z
    #  where M is positive definite and Hermitian, so the result is
    # real valued:
    # https://en.wikipedia.org/wiki/Definiteness_of_a_matrix#Definitions_for_complex_matrices
    result = _vdot(x.real, y.real)
    if jnp.iscomplexobj(x) or jnp.iscomplexobj(y):
        result += _vdot(x.imag, y.imag)
    return result


def tree_vdot(x, y):
    z = jax.tree.leaves(jax.tree.map(_vdot, x, y))
    return sum(z[1:], z[0])


def tree_vdot_real_part(x, y):
    z = jax.tree.leaves(jax.tree.map(_vdot_real_part, x, y))
    return sum(z[1:], z[0])


def tree_add(x, y):
    return jax.tree.map(jax.lax.add, x, y)


def tree_scalar_mul(alpha, x):
    return jax.tree.map(lambda x: alpha * x, x)


def tree_neg(x):
    return jax.tree.map(jax.lax.neg, x)


def tree_sub(x, y):
    return jax.tree.map(jax.lax.sub, x, y)


def hvp_linearized(f, params):
    # Compute the gradient function and linearize it at params
    grad_f = jax.grad(f)
    _, jvp_lin = jax.linearize(grad_f, params)
    # lin_fun is a function that computes the JVP of grad_f at params
    return jvp_lin  # This function computes HVPs for different v


def hvp_forward_over_reverse(f, params):
    def hvp(v):
        return jax.jvp(jax.grad(f), (params,), (v,))[1]

    return hvp


def hvp_reverse_over_reverse(f, params):
    def hvp(v):
        return jax.grad(lambda y: jnp.vdot(jax.grad(f)(y), v))(params)

    return hvp


def hvp_reverse_over_forward(f, params):
    def hvp(v):
        jvp_fun = lambda params: jax.jvp(f, (params,), (v,))[1]
        return jax.grad(jvp_fun)(params)

    return hvp


def grad_and_hvp(f, params, v):
    """
    Compute the gradient and Hessian-vector product of a function.

    Args:
        f: the function to differentiate, should be scalar output
        params: the parameters to differentiate with respect to
        v: the vector to multiply the Hessian with

    Returns:
        the gradient and Hessian-vector product
    """
    return jax.jvp(jax.grad(f), (params,), (v,))


def build_hvp(f, params, linearise: bool = True):
    """
    Build a function that computes the Hessian-vector product of a function.

    Args:
        f: scalar function to differentiate
        params: the parameters to differentiate with respect to
        linearise: whether to linearize the gradient function at params, can be better for reapplying the HVP multiple
            times.

    Returns:
        a function that computes the Hessian-vector product
    """
    if linearise:
        # Compute the gradient function and linearize it at params
        grad_f = jax.grad(f)
        # lin_fun is a function that computes the JVP of grad_f at params
        _, grad_jvp_lin = jax.linearize(grad_f, params)

        def matvec(v):
            return grad_jvp_lin(v)
    else:
        def matvec(v):
            return grad_and_hvp(f, params, v)[1]
    return matvec



from typing import NamedTuple, Optional, List, Union

from jaxns.internals.types import FloatArray, IntArray, XType, UType, MeasureType, BoolArray, PRNGKey

__all__ = [
    'TerminationCondition',
    'NestedSamplerResults',
    'NestedSamplerState'
]


class EvidenceCalculation(NamedTuple):
    """
    Contains a running estimate of evidence and related quantities.
    """
    log_L: FloatArray
    log_X_mean: FloatArray
    log_X2_mean: FloatArray
    log_Z_mean: FloatArray
    log_ZX_mean: FloatArray
    log_Z2_mean: FloatArray
    log_dZ_mean: FloatArray
    log_dZ2_mean: FloatArray


class TerminationCondition(NamedTuple):
    """
    Contains the termination conditions for the nested sampling run.

    Args:
        ess: The effective sample size, if the ESS (Kish's estimate) is greater than this the run will terminate.
        evidence_uncert: The uncertainty in the evidence, if the uncertainty is less than this the run will terminate.
        live_evidence_frac: Depreceated use dlogZ.
        dlogZ: Terminate if log(Z_current + Z_remaining) - log(Z_current) < dlogZ. Default log(1 + 1e-2)
        max_samples: Terminate if the number of samples exceeds this.
        max_num_likelihood_evaluations: Terminate if the number of likelihood evaluations exceeds this.
        log_L_contour: Terminate if this log(L) contour is reached. A contour is reached if any dead point
            has log(L) > log_L_contour. Uncollected live points are not considered.
        efficiency_threshold: Terminate if the efficiency (num_samples / num_likelihood_evaluations) is less than this,
            for the last shrinkage iteration.
        rtol: finish when the relative value 2*|log_L_max - log_L_min|/|log_L_max + log_L_min| < rol
        atol: finish when the absolute |log_L_max - log_L_min| < atol
    """
    ess: Optional[Union[FloatArray, IntArray]] = None
    evidence_uncert: Optional[FloatArray] = None
    live_evidence_frac: Optional[FloatArray] = None  # Depreceated use dlogZ
    dlogZ: Optional[FloatArray] = None  #
    max_samples: Optional[Union[FloatArray, IntArray]] = None
    max_num_likelihood_evaluations: Optional[Union[FloatArray, IntArray]] = None
    log_L_contour: Optional[FloatArray] = None
    efficiency_threshold: Optional[FloatArray] = None
    rtol: Optional[FloatArray] = None
    atol: Optional[FloatArray] = None
    peak_XL_frac: Optional[FloatArray] = None

    def __and__(self, other):
        return TerminationConditionConjunction(conds=[self, other])

    def __or__(self, other):
        return TerminationConditionDisjunction(conds=[self, other])


class TerminationConditionConjunction(NamedTuple):
    """
    A conjunction of termination conditions, e.g. term_cond1 & term_cond2
    """
    conds: List[Union['TerminationConditionDisjunction', 'TerminationConditionConjunction', TerminationCondition]]


class TerminationConditionDisjunction(NamedTuple):
    """
    A disjunction of termination conditions, e.g. term_cond1 | term_cond2
    """
    conds: List[Union['TerminationConditionDisjunction', TerminationConditionConjunction, TerminationCondition]]


class NestedSamplerResults(NamedTuple):
    """
    Results of the nested sampling run.
    """
    log_Z_mean: FloatArray  # estimate of E[log(Z)]
    log_Z_uncert: FloatArray  # estimate of StdDev[log(Z)]
    ESS: FloatArray  # estimate of Kish's effective sample size
    H_mean: FloatArray  # estimate of E[int log(L) L dp/Z]
    samples: XType  # Dict of arrays with leading dimension num_samples
    parametrised_samples: XType  # Dict of arrays with leading dimension num_samples
    U_samples: UType  # Dict of arrays with leading dimension num_samples
    log_L_samples: FloatArray  # log(L) of each sample
    log_dp_mean: FloatArray  # log(E[dZ]) of each sample, where dZ is how much it contributes to the total evidence.
    log_X_mean: FloatArray  # log(E[U]) of each sample
    log_posterior_density: FloatArray  # log(P( theta | D )) log posteriori density
    num_live_points_per_sample: IntArray  # how many live points were taken for the samples.
    num_likelihood_evaluations_per_sample: IntArray  # how many likelihood evaluations were made per sample.
    total_num_samples: IntArray  # int, the total number of samples collected.
    total_phantom_samples: IntArray  # int, the total number of phantom samples collected.
    total_num_likelihood_evaluations: IntArray  # how many likelihood evaluations were made in total
    log_efficiency: FloatArray  # log(total_num_samples / total_num_likelihood_evaluations)
    termination_reason: IntArray  # this will be an int reflecting the reason for termination


class Sample(NamedTuple):
    U_sample: UType  # [..., D] sample in U-space
    log_L_constraint: FloatArray  # [...,] log(L) constraint
    log_L: FloatArray  # [...,] log(L) of sample
    num_likelihood_evaluations: IntArray  # [...,] number of likelihood evaluations


class LivePointCollection(NamedTuple):
    sender_node_idx: IntArray  # [N] with values in [0, N], 0 means root node
    U_sample: UType  # [N, D] sample in U-space
    log_L_constraint: FloatArray  # [N,] log(L) constraint
    log_L: FloatArray  # [N] log(L) of sample
    num_likelihood_evaluations: IntArray  # [N] number of likelihood evaluations


class SampleCollection(NamedTuple):
    sender_node_idx: IntArray  # [N] with values in [0, N], 0 means root node
    log_L: MeasureType  # [N] log(L) of each sample
    U_samples: UType  # [N, D] samples in U-space
    num_likelihood_evaluations: IntArray  # [N] number of likelihood evaluations for each sample
    phantom: BoolArray  # [N] whether the sample is a phantom sample


StaticStandardSampleCollection = SampleCollection


class TerminationRegister(NamedTuple):
    num_samples_used: IntArray
    evidence_calc: EvidenceCalculation
    evidence_calc_with_remaining: EvidenceCalculation
    num_likelihood_evaluations: IntArray
    log_L_contour: FloatArray
    efficiency: FloatArray
    plateau: BoolArray
    no_seed_points: BoolArray
    relative_spread: FloatArray
    absolute_spread: FloatArray
    peak_log_XL: FloatArray


class NestedSamplerState(NamedTuple):
    key: PRNGKey
    next_sample_idx: IntArray  # the next sample insert index <==> the number of samples
    num_samples: IntArray
    sample_collection: StaticStandardSampleCollection


StaticStandardNestedSamplerState = NestedSamplerState



from jaxns.nested_samplers.sharded.sharded_static import round_up_num_live_points


def test_round_up_num_live_points():
    assert round_up_num_live_points(10, 0.5, 1) == 10
    assert round_up_num_live_points(10, 0.5, 2) == 12
    assert round_up_num_live_points(10, 0.5, 3) == 12



import dataclasses
import io
from typing import NamedTuple, Optional, Union, TextIO, Tuple

import jax.numpy as jnp
import numpy as np
import pylab as plt
from jaxlib import xla_client

from jaxns.experimental.solvers.gauss_newton_cg import newton_cg_solver
from jaxns.framework.bases import BaseAbstractModel
from jaxns.internals.constraint_bijections import quick_unit, quick_unit_inverse
from jaxns.internals.mixed_precision import mp_policy
from jaxns.internals.types import PRNGKey, IntArray, UType, FloatArray, LikelihoodInputType, XType
from jaxns.nested_samplers import ShardedStaticNestedSampler
from jaxns.nested_samplers.common.types import TerminationCondition, SampleCollection
from jaxns.samplers.abc import AbstractSampler
from jaxns.utils import _bit_mask

__all__ = [
    'GlobalOptimisationResults',
    'GlobalOptimisationTerminationCondition',
    'GlobalOptimisationState',
    'SimpleGlobalOptimisation'
]


class GlobalOptimisationState(NamedTuple):
    key: PRNGKey
    samples: SampleCollection
    num_samples: IntArray
    relative_spread: FloatArray
    absolute_spread: FloatArray
    num_likelihood_evaluations: IntArray


class GlobalOptimisationResults(NamedTuple):
    U_solution: UType
    X_solution: XType
    solution: LikelihoodInputType
    log_L_solution: FloatArray
    log_L_progress: FloatArray
    num_likelihood_evaluations: IntArray
    num_samples: IntArray
    termination_reason: IntArray
    relative_spread: FloatArray
    absolute_spread: FloatArray


class GlobalOptimisationTerminationCondition(NamedTuple):
    max_likelihood_evaluations: Optional[IntArray] = None  # jnp.asarray(jnp.iinfo(int_type).max, int_type)
    log_likelihood_contour: Optional[FloatArray] = None  # jnp.asarray(jnp.finfo(float_type).max, float_type)
    rtol: Optional[FloatArray] = None  # jnp.asarray(0., float_type)
    atol: Optional[FloatArray] = None  # jnp.asarray(0., float_type)
    min_efficiency: Optional[FloatArray] = None  # jnp.asarray(0., float_type)


def gradient_based_optimisation(model: BaseAbstractModel, init_U_point: UType) -> Tuple[UType, FloatArray, IntArray]:
    def loss(U_unconstrained: UType):
        U = quick_unit(U_unconstrained)
        return -model.log_prob_likelihood(U, allow_nan=False)

    solution, diagnostics = newton_cg_solver(
        loss, quick_unit_inverse(init_U_point)
    )
    final_iter = jnp.max(diagnostics.iteration)
    final_obj = -diagnostics.f[final_iter]
    # Assuming ~4 function evaluations per CG iteration:
    # 1 for the initial function evaluation, 3 for the gradient and hvp.
    num_fun_eval = 4 * jnp.sum(diagnostics.cg_iters)

    return quick_unit(solution), final_obj, num_fun_eval


@dataclasses.dataclass(eq=False)
class SimpleGlobalOptimisation:
    """
    Simple global optimisation leveraging building blocks of nested sampling.
    """
    sampler: AbstractSampler
    num_search_chains: int
    model: BaseAbstractModel
    shell_frac: float = 0.5
    devices: Optional[xla_client.Device] = None
    verbose: bool = False

    def __post_init__(self):
        if self.num_search_chains < 1:
            raise ValueError("num_search_chains must be >= 1.")
        self.num_search_chains = int(self.num_search_chains)

        self._nested_sampler = ShardedStaticNestedSampler(
            model=self.model,
            max_samples=self.num_search_chains * 10,
            init_efficiency_threshold=0.1,
            sampler=self.sampler,
            num_live_points=self.num_search_chains,
            shell_fraction=self.shell_frac,
            devices=self.devices,
            verbose=self.verbose
        )

    def _gradient_descent(self, results: GlobalOptimisationResults) -> GlobalOptimisationResults:
        U_solution, log_L_solution, _num_likelihood_evals = gradient_based_optimisation(self.model,
                                                                                        init_U_point=results.U_solution)
        X_solution = self.model.transform(U_solution)
        solution = self.model.prepare_input(U_solution)
        num_likelihood_evals = results.num_likelihood_evaluations + _num_likelihood_evals
        return results._replace(
            U_solution=U_solution,
            log_L_solution=log_L_solution,
            X_solution=X_solution,
            solution=solution,
            num_likelihood_evaluations=num_likelihood_evals
        )

    def _to_results(self, termination_reason: IntArray, state: GlobalOptimisationState) -> GlobalOptimisationResults:
        """
        Converts the final state of the global optimisation to results.

        Args:
            state: final state of the global optimisation

        Returns:
            results of the global optimisation
        """
        is_sample_mask = jnp.arange(np.shape(state.samples.log_L)[0], dtype=mp_policy.index_dtype) < state.num_samples
        log_L_masked = jnp.where(is_sample_mask, state.samples.log_L, jnp.asarray(jnp.nan, mp_policy.measure_dtype))
        best_idx = jnp.nanargmax(log_L_masked)
        U_solution = state.samples.U_samples[best_idx]
        X_solution = self.model.transform(U_solution)
        solution = self.model.prepare_input(U_solution)  # The output of prior_model is solution
        log_L_solution = state.samples.log_L[best_idx]
        num_likelihood_evaluations = state.num_likelihood_evaluations
        log_L_progress = jnp.sort(log_L_masked)  # Low to high likelihoods, nan's at the end.
        return GlobalOptimisationResults(
            U_solution=U_solution,
            X_solution=X_solution,
            solution=solution,
            log_L_solution=log_L_solution,
            log_L_progress=log_L_progress,
            num_likelihood_evaluations=num_likelihood_evaluations,
            num_samples=state.num_samples,
            relative_spread=state.relative_spread,
            absolute_spread=state.absolute_spread,
            termination_reason=termination_reason
        )

    def _run(self, key: PRNGKey, term_cond: GlobalOptimisationTerminationCondition) -> Tuple[
        IntArray, GlobalOptimisationState]:
        """
        Runs the global optimisation.

        Args:
            key: PRNGKey
            term_cond: termination condition

        Returns:
            the termination reason and final state of the global optimisation
        """

        termination_reason, termination_register, state = self._nested_sampler._run(
            key=key,
            term_cond=TerminationCondition(
                max_num_likelihood_evaluations=term_cond.max_likelihood_evaluations,
                log_L_contour=term_cond.log_likelihood_contour,
                efficiency_threshold=term_cond.min_efficiency,
                atol=term_cond.atol,
                rtol=term_cond.rtol,
                max_samples=None  # Turn off max samples for global optimisation, wraps index
            )
        )

        go_state = GlobalOptimisationState(
            key=state.key,
            samples=state.sample_collection,
            num_samples=state.num_samples,
            absolute_spread=termination_register.absolute_spread,
            relative_spread=termination_register.relative_spread,
            num_likelihood_evaluations=termination_register.num_likelihood_evaluations
        )
        return termination_reason, go_state


def plot_progress(results: GlobalOptimisationResults, save_file: Optional[str] = None):
    fig, ax = plt.subplots(1, 1, figsize=(8, 6))
    start_sample_idx = max(0, results.num_samples - int(np.shape(results.log_L_progress)[0]))
    end_sample_idx = int(results.num_samples)

    ax.plot(
        np.arange(start_sample_idx, end_sample_idx),
        results.log_L_progress
    )
    ax.set_title(f"Optimisation progress (samples {start_sample_idx} to {end_sample_idx})")
    ax.set_xlabel("Sample index")
    ax.set_ylabel(r"Objective ($\log L$)")
    if save_file is not None:
        plt.savefig(save_file)
    plt.show()


def go_summary(results: GlobalOptimisationResults, f_obj: Optional[Union[str, TextIO]] = None):
    """
    Gives a summary of the results of a global optimisation.

    Args:
        results (GlobalOptimisationResults): Nested sampler result
    """
    main_s = []

    def _print(s):
        print(s)
        main_s.append(s)

    def _round(v, uncert_v):
        v = float(v)
        uncert_v = float(uncert_v)
        try:
            sig_figs = -int("{:e}".format(uncert_v).split('e')[1]) + 1
            return round(float(v), sig_figs)
        except:
            return float(v)

    def _print_termination_condition(_termination_reason: int):
        termination_bit_mask = _bit_mask(int(_termination_reason), width=11)
        for bit, condition in zip(termination_bit_mask, [
            'Reached max samples',
            'Evidence uncertainty low enough',
            'Small remaining evidence',
            'Reached ESS',
            "Used max num likelihood evaluations",
            'Likelihood contour reached',
            'Sampler efficiency too low',
            'All live-points are on a single plateau (sign of possible precision error)',
            'relative spread of live points < rtol',
            'absolute spread of live points < atol',
            'no seed points left (consider decreasing shell_fraction)'
        ]):
            if bit == 1:
                _print(condition)

    _print("--------")
    _print("Termination Conditions:")
    if np.size(results.termination_reason) > 1:
        for replica_idx in range(np.size(results.termination_reason)):
            _print(f"Replica {replica_idx}:")
            _print_termination_condition(int(results.termination_reason[replica_idx]))
    else:
        _print_termination_condition(int(results.termination_reason))
    _print("--------")
    _print(f"likelihood evals: {int(results.num_likelihood_evaluations):d}")
    _print(f"samples: {int(results.num_samples):d}")
    _print(
        f"likelihood evals / sample: {float(results.num_likelihood_evaluations / results.num_samples):.1f}"
    )
    _print("--------")
    _print(
        f"max(log_L)={_round(results.log_L_solution, results.log_L_solution)}"
    )
    _print(
        f"relative spread: {_round(results.relative_spread, results.relative_spread)}"
    )
    _print(
        f"absolute spread: {_round(results.absolute_spread, results.absolute_spread)}"
    )

    X_solution = results.X_solution
    for name in X_solution.keys():

        # For shaped arrays, we want to print "{name}[{i0, i1, ..., in}]" for all valid indices.

        if np.size(X_solution[name]) == 0:
            continue

        _print("--------")

        shape = np.shape(X_solution[name])
        num_dims = len(shape)
        is_shaped = num_dims > 0

        var_name = f"{name}[{','.join(['#'] * num_dims)}]" if is_shaped else name
        _print(
            f"{var_name}: max(L) est."
        )
        if is_shaped:
            indices = np.indices(shape).reshape((num_dims, -1)).T

            for inds in indices:
                _max_like_point = X_solution[name][tuple(inds)]
                _print(f"{name}[{','.join(str(i) for i in inds)}]: {_round(_max_like_point, 0.1 * _max_like_point)}")
        else:
            _max_like_point = X_solution[name]
            _print(f"{name}: {_round(_max_like_point, 0.1 * _max_like_point)}")

    _print("--------")
    if f_obj is not None:
        out = "\n".join(main_s)
        if isinstance(f_obj, str):
            with open(f_obj, 'w') as f:
                f.write(out)
        elif isinstance(f_obj, io.TextIOBase):
            f_obj.write(out)
        else:
            raise TypeError(f"Invalid f_obj: {type(f_obj)}")



from typing import NamedTuple

import jax
import jax.numpy as jnp
import numpy as np

from jaxns.internals.namedtuple_utils import isinstance_namedtuple
from jaxns.internals.namedtuple_utils import serialise_namedtuple, \
    deserialise_namedtuple, issubclass_namedtuple, serialise_ndarray


class MockNested(NamedTuple):
    name: str


class MockModel(NamedTuple):
    string: str
    integer: int
    float: float
    complex: complex
    ndarray: np.ndarray
    jaxndarray: jax.Array
    nested: MockNested


def test_isinstance_namedtuple():
    # Example NamedTuple
    data = MockModel(
        string='Alice',
        integer=25,
        float=3.14,
        complex=1 + 2j,
        ndarray=np.array([1, 2, 3]),
        jaxndarray=jnp.array([1, 2, 3]),
        nested=MockNested('Bob')
    )

    assert isinstance_namedtuple(data)
    assert not isinstance_namedtuple(('Bob',))

    assert not isinstance(('Bob',), MockNested)


def test_issubclass_namedtuple():
    class TestSubclass(NamedTuple):
        pass

    assert issubclass_namedtuple(TestSubclass)

    class TestSubclass(tuple):
        pass

    assert not issubclass_namedtuple(TestSubclass)

    class C:
        pass

    assert not issubclass_namedtuple(C)


def test_serialise_namedtuple():
    data = MockModel(
        string='Alice',
        integer=25,
        float=3.14,
        complex=1 + 2j,
        ndarray=np.array([1, 2, 3]),
        jaxndarray=jnp.array([1, 2, 3]),
        nested=MockNested('Bob')
    )

    # Serialise
    serialized_data = serialise_namedtuple(data)
    print(serialized_data)

    # Deserialize
    restored_data = deserialise_namedtuple(serialized_data)
    print(restored_data)

    assert data.string == restored_data.string
    assert data.integer == restored_data.integer
    assert data.float == restored_data.float
    assert data.complex == restored_data.complex
    np.testing.assert_allclose(data.ndarray, restored_data.ndarray)
    np.testing.assert_allclose(data.jaxndarray, restored_data.jaxndarray)
    assert data.nested.name == restored_data.nested.name


def test_serialise_ndarray():
    array = np.array([1, 2, 3])
    serialized_data = serialise_ndarray(array)
    print(serialized_data)
    restored_data = deserialise_namedtuple(serialized_data)
    print(restored_data)

    np.testing.assert_allclose(array, restored_data)


def test_serialise_jax_ndarray():
    array = jnp.array([1, 2, 3])
    serialized_data = serialise_ndarray(array)
    print(serialized_data)
    restored_data = deserialise_namedtuple(serialized_data)
    print(restored_data)

    np.testing.assert_allclose(array, restored_data)
